from collections import defaultdict
from itertools import islice
from pathlib import Path, PurePath, PurePosixPath
from typing import Any, List, Optional, Set, Tuple, Union

import anndata as ad
import fsspec
import lamindb_setup
import pandas as pd
from anndata import AnnData
from lamin_utils import colors, logger
from lamindb_setup import settings as setup_settings
from lamindb_setup._init_instance import register_storage
from lamindb_setup.dev import StorageSettings
from lamindb_setup.dev._docs import doc_args
from lamindb_setup.dev._hub_utils import get_storage_region
from lamindb_setup.dev.upath import create_path
from lnschema_core import Feature, FeatureSet, File, Modality, Run, Storage
from lnschema_core.types import AnnDataLike, DataLike, FieldAttr, PathLike

from lamindb._utils import attach_func_to_class_method
from lamindb.dev._data import _track_run_input
from lamindb.dev._settings import settings
from lamindb.dev.hashing import b16_to_b64, hash_file
from lamindb.dev.storage import (
    LocalPathClasses,
    UPath,
    delete_storage,
    infer_suffix,
    load_to_memory,
    size_adata,
    write_to_file,
)
from lamindb.dev.storage._backed_access import AnnDataAccessor, BackedAccessor
from lamindb.dev.storage.file import (
    ProgressCallback,
    auto_storage_key_from_file,
    auto_storage_key_from_id_suffix,
    extract_suffix_from_path,
    filepath_from_file,
)
from lamindb.dev.versioning import get_ids_from_old_version, init_id

from . import _TESTING
from ._feature import convert_numpy_dtype_to_lamin_feature_type
from .dev._data import (
    add_transform_to_kwargs,
    get_run,
    save_feature_set_links,
    save_transform_run_feature_sets,
)
from .dev._priors import priors
from .dev.storage.file import AUTO_KEY_PREFIX


def process_pathlike(
    filepath: UPath, skip_existence_check: bool = False
) -> Tuple[Storage, bool]:
    if not skip_existence_check:
        try:  # check if file exists
            if not filepath.exists():
                raise FileNotFoundError(filepath)
        except PermissionError:
            pass
    if isinstance(filepath, LocalPathClasses):
        filepath = filepath.resolve()
    # check whether the path is in default storage
    default_storage = lamindb_setup.settings.storage.record
    if check_path_is_child_of_root(filepath, default_storage.root_as_path()):
        use_existing_storage_key = True
        return default_storage, use_existing_storage_key
    else:
        # check whether the path is part of one of the existing
        # already-registered storage locations
        result = check_path_in_existing_storage(filepath)
        if isinstance(result, Storage):
            use_existing_storage_key = True
            return result, use_existing_storage_key
        else:
            # if the path is in the cloud, we have a good candidate
            # for the storage root: the bucket
            if not isinstance(filepath, LocalPathClasses):
                # for a cloud path, new_root is always the bucket name
                new_root = list(filepath.parents)[-1]
                new_root_str = new_root.as_posix().rstrip("/")
                region = get_storage_region(new_root_str)
                storage_settings = StorageSettings(new_root_str, region)
                register_storage(storage_settings)
                use_existing_storage_key = True
                return storage_settings.record, use_existing_storage_key
            # if the filepath is local
            else:
                use_existing_storage_key = False
                # if the default storage is local we'll throw an error if the user
                # doesn't provide a key
                if not lamindb_setup.settings.storage.is_cloud:
                    return default_storage, use_existing_storage_key
                # if the default storage is in the cloud (the file is going to
                # be uploaded upon saving it), we treat the filepath as a cache
                else:
                    return default_storage, use_existing_storage_key


def process_data(
    provisional_id: str,
    data: Union[PathLike, DataLike],
    format: Optional[str],
    key: Optional[str],
    skip_existence_check: bool = False,
) -> Tuple[Any, Union[Path, UPath], str, Storage, bool]:
    """Serialize a data object that's provided as file or in memory."""
    # if not overwritten, data gets stored in default storage
    if isinstance(data, (str, Path, UPath)):  # PathLike, spelled out
        filepath = create_path(data)
        storage, use_existing_storage_key = process_pathlike(
            filepath, skip_existence_check=skip_existence_check
        )
        suffix = extract_suffix_from_path(filepath)
        memory_rep = None
    elif isinstance(data, (pd.DataFrame, AnnData)):  # DataLike, spelled out
        storage = lamindb_setup.settings.storage.record
        memory_rep = data
        if key is not None:
            key_suffix = extract_suffix_from_path(PurePosixPath(key), arg_name="key")
            # use suffix as the (adata) format if the format is not provided
            if isinstance(data, AnnData) and format is None and len(key_suffix) > 0:
                format = key_suffix[1:]
        else:
            key_suffix = None
        suffix = infer_suffix(data, format)
        if key_suffix is not None and key_suffix != suffix:
            raise ValueError(
                f"The suffix '{key_suffix}' of the provided key is incorrect, it should"
                f" be '{suffix}'."
            )
        cache_name = f"{provisional_id}{suffix}"
        filepath = lamindb_setup.settings.storage.cache_dir / cache_name
        # Alex: I don't understand the line below
        if filepath.suffixes == []:
            filepath = filepath.with_suffix(suffix)
        if suffix not in {".zarr", ".zrad"}:
            write_to_file(data, filepath)
        use_existing_storage_key = False
    else:
        raise NotImplementedError(
            f"Do not know how to create a file object from {data}, pass a filepath"
            " instead!"
        )
    return memory_rep, filepath, suffix, storage, use_existing_storage_key


def get_hash(
    filepath: UPath,
    suffix,
    filepath_stat=None,
    check_hash: bool = True,
) -> Union[Tuple[Optional[str], Optional[str]], File]:
    if suffix in {".zarr", ".zrad"}:
        return None
    if not isinstance(filepath, LocalPathClasses):
        stat = filepath_stat
        if stat is not None and "ETag" in stat:
            # small files
            if "-" not in stat["ETag"]:
                # only store hash for non-multipart uploads
                # we can't rapidly validate multi-part uploaded files client-side
                # we can add more logic later down-the-road
                hash = b16_to_b64(stat["ETag"])
                hash_type = "md5"
            else:
                stripped_etag, suffix = stat["ETag"].split("-")
                suffix = suffix.strip('"')
                hash = f"{b16_to_b64(stripped_etag)}-{suffix}"
                hash_type = "md5-n"  # this is the S3 chunk-hashing strategy
        else:
            logger.warning(f"did not add hash for {filepath}")
            return None, None
    else:
        hash, hash_type = hash_file(filepath)
    if not check_hash:
        return hash, hash_type
    result = File.filter(hash=hash).list()
    if len(result) > 0:
        if settings.upon_file_create_if_hash_exists == "error":
            msg = f"file with same hash exists: {result[0]}"
            hint = (
                "💡 you can make this error a warning:\n"
                "    ln.settings.upon_file_create_if_hash_exists"
            )
            raise RuntimeError(f"{msg}\n{hint}")
        elif settings.upon_file_create_if_hash_exists == "warn_create_new":
            logger.warning(
                "creating new File object despite existing file with same hash:"
                f" {result[0]}"
            )
            return hash, hash_type
        else:
            logger.warning(f"returning existing file with same hash: {result[0]}")
            return result[0]
    else:
        return hash, hash_type


def get_path_size_hash(
    filepath: UPath,
    memory_rep: Optional[Union[pd.DataFrame, AnnData]],
    suffix: str,
    check_hash: bool = True,
):
    cloudpath = None
    localpath = None
    hash_and_type: Tuple[Optional[str], Optional[str]]

    if suffix in {".zarr", ".zrad"}:
        if memory_rep is not None:
            size = size_adata(memory_rep)
        else:
            if not isinstance(filepath, LocalPathClasses):
                cloudpath = filepath
                # todo: properly calculate size
                size = 0
            else:
                localpath = filepath
                size = sum(
                    f.stat().st_size for f in filepath.rglob("*") if f.is_file()  # type: ignore # noqa
                )
        hash_and_type = None, None
    else:
        # to accelerate ingesting high numbers of files
        if settings.upon_file_create_skip_size_hash:
            size = None
            hash_and_type = None, None
        else:
            filepath_stat = filepath.stat()
            if not isinstance(filepath, LocalPathClasses):
                size = filepath_stat["size"]
                cloudpath = filepath
                hash_and_type = None, None
            else:
                size = filepath_stat.st_size  # type: ignore
                localpath = filepath
            hash_and_type = get_hash(
                filepath, suffix, filepath_stat=filepath_stat, check_hash=check_hash
            )
    return localpath, cloudpath, size, hash_and_type


def check_path_in_existing_storage(
    filepath: Union[Path, UPath]
) -> Union[Storage, bool]:
    for storage in Storage.filter().all():
        # if path is part of storage, return it
        if check_path_is_child_of_root(filepath, root=create_path(storage.root)):
            return storage
    return False


def check_path_is_child_of_root(
    filepath: Union[Path, UPath], root: Optional[Union[Path, UPath]] = None
) -> bool:
    if root is None:
        root = lamindb_setup.settings.storage.root

    filepath = UPath(str(filepath)) if not isinstance(filepath, UPath) else filepath
    root = UPath(str(root)) if not isinstance(root, UPath) else root

    # the following comparisons can fail if types aren't comparable
    if not isinstance(filepath, LocalPathClasses) and not isinstance(
        root, LocalPathClasses
    ):
        # the following tests equivalency of two UPath objects
        # via string representations; otherwise
        # S3Path('s3://lndb-storage/') and S3Path('s3://lamindb-ci/')
        # test as equivalent
        return list(filepath.parents)[-1].as_posix() == root.as_posix()
    elif isinstance(filepath, LocalPathClasses) and isinstance(root, LocalPathClasses):
        return root.resolve() in filepath.resolve().parents
    else:
        return False


def get_relative_path_to_directory(
    path: Union[PurePath, Path, UPath], directory: Union[PurePath, Path, UPath]
) -> Union[PurePath, Path]:
    if isinstance(directory, UPath) and not isinstance(directory, LocalPathClasses):
        # UPath.relative_to() is not behaving as it should (2023-04-07)
        # need to lstrip otherwise inconsistent behavior across trailing slashes
        # see test_file.py: test_get_relative_path_to_directory
        relpath = PurePath(
            path.as_posix().replace(directory.as_posix(), "").lstrip("/")
        )
    elif isinstance(directory, Path):
        relpath = path.resolve().relative_to(directory.resolve())  # type: ignore
    elif isinstance(directory, PurePath):
        relpath = path.relative_to(directory)
    else:
        raise TypeError("Directory not of type Path or UPath")
    return relpath


def get_file_kwargs_from_data(
    *,
    data: Union[Path, UPath, str, pd.DataFrame, AnnData],
    key: Optional[str],
    run: Optional[Run],
    format: Optional[str],
    provisional_id: str,
    skip_check_exists: bool = False,
):
    run = get_run(run)
    memory_rep, filepath, suffix, storage, use_existing_storage_key = process_data(
        provisional_id, data, format, key, skip_check_exists
    )
    # the following will return a localpath that is not None if filepath is local
    # it will return a cloudpath that is not None if filepath is on the cloud
    local_filepath, cloud_filepath, size, hash_and_type = get_path_size_hash(
        filepath,
        memory_rep,
        suffix,
    )
    if isinstance(hash_and_type, File):
        return hash_and_type, None
    else:
        hash, hash_type = hash_and_type

    check_path_in_storage = False
    if use_existing_storage_key:
        inferred_key = get_relative_path_to_directory(
            path=filepath, directory=storage.root_as_path()
        ).as_posix()
        if key is None:
            key = inferred_key
        else:
            if not key == inferred_key:
                raise ValueError(
                    f"The path '{data}' is already in registered storage"
                    f" '{storage.root}' with key '{inferred_key}'\nYou passed"
                    f" conflicting key '{key}': please move the file before"
                    " registering it."
                )
        check_path_in_storage = True
    else:
        storage = lamindb_setup.settings.storage.record

    if key is not None and key.startswith(AUTO_KEY_PREFIX):
        raise ValueError(f"Key cannot start with {AUTO_KEY_PREFIX}")

    log_storage_hint(
        check_path_in_storage=check_path_in_storage,
        storage=storage,
        key=key,
        id=provisional_id,
        suffix=suffix,
    )

    kwargs = dict(
        suffix=suffix,
        hash=hash,
        hash_type=hash_type,
        key=key,
        size=size,
        storage_id=storage.id,
        # passing both the id and the object
        # to make them both available immediately
        # after object creation
        run_id=run.id if run is not None else None,
        run=run,
    )
    privates = dict(
        local_filepath=local_filepath,
        cloud_filepath=cloud_filepath,
        memory_rep=memory_rep,
        check_path_in_storage=check_path_in_storage,
    )

    return kwargs, privates


def log_storage_hint(
    *,
    check_path_in_storage: bool,
    storage: Optional[Storage],
    key: Optional[str],
    id: str,
    suffix: str,
) -> None:
    hint = ""
    if check_path_in_storage:
        display_root = storage.root  # type: ignore
        # check whether path is local
        if fsspec.utils.get_protocol(storage.root) == "file":  # type: ignore
            # if it's a local path, check whether it's in the current working directory
            root_path = Path(storage.root)  # type: ignore
            if check_path_is_child_of_root(root_path, Path.cwd()):
                # only display the relative path, not the fully resolved path
                display_root = root_path.relative_to(Path.cwd())
        hint += f"file in storage '{display_root}'"  # type: ignore
    else:
        hint += "file will be copied to default storage upon `save()`"
    if key is None:
        storage_key = auto_storage_key_from_id_suffix(id, suffix)
        hint += f" with key `None` ('{storage_key}')"
    else:
        hint += f" with key '{key}'"
    logger.hint(hint)


def data_is_anndata(data: DataLike):
    if isinstance(data, AnnData):
        return True
    if isinstance(data, (str, Path, UPath)):
        return Path(data).suffix in {".h5ad", ".zrad"}
    return False  # pragma: no cover


def data_is_mudata(data: DataLike):  # pragma: no cover
    try:
        from mudata import MuData
    except ModuleNotFoundError:
        return False

    if isinstance(data, MuData):
        return True
    if isinstance(data, (str, Path, UPath)):
        return Path(data).suffix in {".h5mu"}
    return False


def __init__(file: File, *args, **kwargs):
    # Below checks for the Django-internal call in from_db()
    # it'd be better if we could avoid this, but not being able to create a File
    # from data with the default constructor renders the central class of the API
    # essentially useless
    # The danger below is not that a user might pass as many args (12 of it), but rather
    # that at some point the Django API might change; on the other hand, this
    # condition of for calling the constructor based on kwargs should always
    # stay robust
    if len(args) == len(file._meta.concrete_fields):
        super(File, file).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: data")
    data: Union[PathLike, DataLike] = kwargs.pop("data") if len(args) == 0 else args[0]
    key: Optional[str] = kwargs.pop("key") if "key" in kwargs else None
    run: Optional[Run] = kwargs.pop("run") if "run" in kwargs else None
    description: Optional[str] = (
        kwargs.pop("description") if "description" in kwargs else None
    )
    is_new_version_of: Optional[File] = (
        kwargs.pop("is_new_version_of") if "is_new_version_of" in kwargs else None
    )
    initial_version_id: Optional[str] = (
        kwargs.pop("initial_version_id") if "initial_version_id" in kwargs else None
    )
    version: Optional[str] = kwargs.pop("version") if "version" in kwargs else None
    format = kwargs.pop("format") if "format" in kwargs else None
    log_hint = kwargs.pop("log_hint") if "log_hint" in kwargs else True
    skip_check_exists = (
        kwargs.pop("skip_check_exists") if "skip_check_exists" in kwargs else False
    )

    if not len(kwargs) == 0:
        raise ValueError(
            "Only data, key, run, description, version, is_new_version_of can be"
            f" passed, you passed: {kwargs}"
        )

    if is_new_version_of is None:
        provisional_id = init_id(version=version, n_full_id=20)
    else:
        if not isinstance(is_new_version_of, File):
            raise TypeError("is_new_version_of has to be of type ln.File")
        provisional_id, initial_version_id, version = get_ids_from_old_version(
            is_new_version_of, version, n_full_id=20
        )
        if description is None:
            description = is_new_version_of.description

    if version is not None:
        if initial_version_id is None:
            logger.info(
                "initializing versioning for this file! create future versions of it"
                " using ln.File(..., is_new_version_of=old_file)"
            )
    kwargs_or_file, privates = get_file_kwargs_from_data(
        data=data,
        key=key,
        run=run,
        format=format,
        provisional_id=provisional_id,
        skip_check_exists=skip_check_exists,
    )

    # an object with the same hash already exists
    if isinstance(kwargs_or_file, File):
        from ._registry import init_self_from_db

        # kwargs_or_file is an existing file
        init_self_from_db(file, kwargs_or_file)
        return None
    else:
        kwargs = kwargs_or_file

    if isinstance(data, pd.DataFrame):
        if log_hint:
            logger.hint(
                "data is a dataframe, consider using .from_df() to link column"
                " names as features"
            )
        kwargs["accessor"] = "DataFrame"
    elif data_is_anndata(data):
        if log_hint:
            logger.hint(
                "data is AnnDataLike, consider using .from_anndata() to link"
                " var_names and obs.columns as features"
            )
        kwargs["accessor"] = "AnnData"
    elif data_is_mudata(data):
        kwargs["accessor"] = "MuData"

    kwargs["id"] = provisional_id
    kwargs["initial_version_id"] = initial_version_id
    kwargs["version"] = version
    kwargs["description"] = description
    # this check needs to come down here because key might be populated from an
    # existing file path during get_file_kwargs_from_data()
    if (
        kwargs["key"] is None
        and kwargs["description"] is None
        and kwargs["run"] is None
    ):
        raise ValueError("Pass one of key, run or description as a parameter")

    add_transform_to_kwargs(kwargs, kwargs["run"])

    if data is not None:
        file._local_filepath = privates["local_filepath"]
        file._cloud_filepath = privates["cloud_filepath"]
        file._memory_rep = privates["memory_rep"]
        file._to_store = not privates["check_path_in_storage"]

    super(File, file).__init__(**kwargs)


@classmethod  # type: ignore
@doc_args(File.from_df.__doc__)
def from_df(
    cls,
    df: "pd.DataFrame",
    field: FieldAttr = Feature.name,
    key: Optional[str] = None,
    description: Optional[str] = None,
    run: Optional[Run] = None,
    modality: Optional[Modality] = None,
) -> "File":
    """{}"""
    file = File(data=df, key=key, run=run, description=description, log_hint=False)
    feature_set = FeatureSet.from_df(df, field=field, modality=modality)
    if feature_set is not None:
        file._feature_sets = {"columns": feature_set}
    else:
        file._feature_sets = {}
    return file


def parse_feature_sets_from_anndata(
    adata: AnnDataLike, field: Optional[FieldAttr], modality: Optional[Modality] = None
):
    data_parse = adata
    if not isinstance(adata, AnnData):  # is a path
        filepath = create_path(adata)  # returns Path for local
        if not isinstance(filepath, LocalPathClasses):
            from lamindb.dev.storage._backed_access import backed_access

            data_parse = backed_access(filepath)
        else:
            data_parse = ad.read(filepath, backed="r")
        type = "float"
    else:
        type = convert_numpy_dtype_to_lamin_feature_type(adata.X.dtype)
    feature_sets = {}
    logger.info("parsing feature names of X stored in slot 'var'")
    logger.indent = "   "
    feature_set_var = FeatureSet.from_values(
        data_parse.var.index,
        field,
        type=type,
        modality=modality,
    )
    if feature_set_var is not None:
        feature_sets["var"] = feature_set_var
        logger.save(f"linked: {feature_set_var}")
    logger.indent = ""
    if len(data_parse.obs.columns) > 0:
        logger.info("parsing feature names of slot 'obs'")
        logger.indent = "   "
        feature_set_obs = FeatureSet.from_df(
            data_parse.obs, modality=priors.modalities.meta
        )
        if feature_set_obs is not None:
            feature_sets["obs"] = feature_set_obs
            logger.save(f"linked: {feature_set_obs}")
        logger.indent = ""
    return feature_sets


@classmethod  # type: ignore
@doc_args(File.from_anndata.__doc__)
def from_anndata(
    cls,
    adata: "AnnDataLike",
    field: Optional[FieldAttr],
    key: Optional[str] = None,
    description: Optional[str] = None,
    run: Optional[Run] = None,
    modality: Optional[Modality] = None,
) -> "File":
    """{}"""
    file = File(data=adata, key=key, run=run, description=description, log_hint=False)
    file._feature_sets = parse_feature_sets_from_anndata(adata, field, modality)
    return file


@classmethod  # type: ignore
@doc_args(File.from_dir.__doc__)
def from_dir(
    cls,
    path: PathLike,
    key: Optional[str] = None,
    *,
    run: Optional[Run] = None,
) -> List["File"]:
    """{}"""
    folderpath: UPath = create_path(path)  # returns Path for local
    storage, use_existing_storage = process_pathlike(folderpath)
    folder_key_path: Union[PurePath, Path]
    if key is None:
        if not use_existing_storage:
            logger.warning(
                "folder is outside existing storage location, will copy files from"
                f" {path} to {storage.root}/{folderpath.name}"
            )
            folder_key_path = Path(folderpath.name)
        else:
            # maintain the hierachy within an existing storage location
            folder_key_path = get_relative_path_to_directory(
                folderpath, storage.root_as_path()
            )
    else:
        folder_key_path = Path(key)

    # always sanitize by stripping a trailing slash
    folder_key = folder_key_path.as_posix().rstrip("/")

    # TODO: (non-local) UPath doesn't list the first level files and dirs with "*"
    pattern = "" if not isinstance(folderpath, LocalPathClasses) else "*"

    # silence fine-grained logging
    verbosity = settings.verbosity
    if verbosity >= 1:
        settings.verbosity = 1  # just warnings
    files = []
    for filepath in folderpath.rglob(pattern):
        if filepath.is_file():
            relative_path = get_relative_path_to_directory(filepath, folderpath)
            file_key = folder_key + "/" + relative_path.as_posix()
            # if creating from rglob, we don't need to check for existence
            file = File(filepath, run=run, key=file_key, skip_check_exists=True)
            files.append(file)
    settings.verbosity = verbosity

    # run sanity check on hashes
    hashes = [file.hash for file in files if file.hash is not None]
    ids = [file.id for file in files]
    if len(set(hashes)) != len(hashes):
        # consider exact duplicates (same id, same hash)
        if len(set(ids)) == len(set(hashes)):
            logger.warning("dropping duplicate records in list of file records")
            files = list(set(files))
        # consider false duplicates (different id, same hash)
        else:
            seen_hashes = set()
            non_unique_files = [
                file
                for file in files
                if file.hash in seen_hashes or seen_hashes.add(file.hash)  # type: ignore  # noqa
            ]
            display_non_unique = "\n    ".join(f"{file}" for file in non_unique_files)
            logger.warning(
                "there are different file ids with the same hashes, dropping"
                f" {len(non_unique_files)} duplicates out of {len(files)} files:\n   "
                f" {display_non_unique}"
            )
            files = [
                file for file in files if file not in set(non_unique_files)
            ]  # noqa
    logger.success(
        f"created {len(files)} files from directory using storage"
        f" {storage.root} and key = {folder_key}/"
    )
    return files


# docstring handled through attach_func_to_class_method
def replace(
    self,
    data: Union[PathLike, DataLike],
    run: Optional[Run] = None,
    format: Optional[str] = None,
) -> None:
    kwargs, privates = get_file_kwargs_from_data(
        provisional_id=self.id,
        data=data,
        key=self.key,
        run=run,
        format=format,
    )
    if self.key is not None:
        key_path = PurePosixPath(self.key)
        new_filename = f"{key_path.stem}{kwargs['suffix']}"
        # the following will only be true if the suffix changes!
        if key_path.name != new_filename:
            self._clear_storagekey = self.key
            self.key = str(key_path.with_name(new_filename))
            logger.warning(
                f"replacing the file will replace key '{key_path}' with '{self.key}'"
                f" and delete '{key_path}' upon `save()`"
            )
    else:
        self.key = kwargs["key"]
        old_storage = auto_storage_key_from_file(self)
        new_storage = (
            self.key if self.key is not None else f"{self.id}{kwargs['suffix']}"
        )
        if old_storage != new_storage:
            self._clear_storagekey = old_storage

    self.suffix = kwargs["suffix"]
    self.size = kwargs["size"]
    self.hash = kwargs["hash"]
    self.run = kwargs["run"]
    self._local_filepath = privates["local_filepath"]
    self._cloud_filepath = privates["cloud_filepath"]
    self._memory_rep = privates["memory_rep"]
    self._to_store = not privates[
        "check_path_in_storage"
    ]  # no need to upload if new file is already in storage


# docstring handled through attach_func_to_class_method
def backed(
    self, is_run_input: Optional[bool] = None
) -> Union["AnnDataAccessor", "BackedAccessor"]:
    suffixes = (".h5", ".hdf5", ".h5ad", ".zrad", ".zarr")
    if self.suffix not in suffixes:
        raise ValueError(
            "File should have a zarr or h5 object as the underlying data, please use"
            " one of the following suffixes for the object name:"
            f" {', '.join(suffixes)}."
        )

    from lamindb.dev.storage._backed_access import backed_access

    _track_run_input(self, is_run_input)

    filepath = filepath_from_file(self)
    # consider the case where an object is already locally cached
    localpath = setup_settings.instance.storage.cloud_to_local_no_update(filepath)
    if localpath.exists():
        return backed_access(localpath)
    else:
        return backed_access(filepath)


# docstring handled through attach_func_to_class_method
def load(
    self, is_run_input: Optional[bool] = None, stream: bool = False, **kwargs
) -> DataLike:
    _track_run_input(self, is_run_input)
    if hasattr(self, "_memory_rep") and self._memory_rep is not None:
        return self._memory_rep
    return load_to_memory(filepath_from_file(self), stream=stream, **kwargs)


# docstring handled through attach_func_to_class_method
def stage(self, is_run_input: Optional[bool] = None) -> Path:
    if self.suffix in {".zrad", ".zarr"}:
        raise RuntimeError("zarr object can't be staged, please use load() or stream()")
    _track_run_input(self, is_run_input)

    filepath = filepath_from_file(self)
    cb = ProgressCallback("downloading")
    return setup_settings.instance.storage.cloud_to_local(filepath, callback=cb)


# docstring handled through attach_func_to_class_method
def delete(self, storage: Optional[bool] = None) -> None:
    if storage is None:
        response = input(f"Are you sure you want to delete {self} from storage? (y/n)")
        delete_in_storage = response == "y"
    else:
        delete_in_storage = storage

    # need to grab file path before deletion
    filepath = self.path
    # only delete in storage if DB delete is successful
    # DB delete might error because of a foreign key constraint violated etc.
    self._delete_skip_storage()
    # we don't yet have any way to bring back the deleted metadata record
    # in case the storage deletion fails - this is important for ACID down the road
    if delete_in_storage:
        delete_storage(filepath)
        logger.success(f"deleted stored object {colors.yellow(f'{filepath}')}")


def _delete_skip_storage(file, *args, **kwargs) -> None:
    super(File, file).delete(*args, **kwargs)


# docstring handled through attach_func_to_class_method
def save(self, *args, **kwargs) -> None:
    self._save_skip_storage(*args, **kwargs)
    from lamindb._save import check_and_attempt_clearing, check_and_attempt_upload

    exception = check_and_attempt_upload(self)
    if exception is not None:
        self._delete_skip_storage()
        raise RuntimeError(exception)
    exception = check_and_attempt_clearing(self)
    if exception is not None:
        raise RuntimeError(exception)


def _save_skip_storage(file, *args, **kwargs) -> None:
    save_transform_run_feature_sets(file)
    super(File, file).save(*args, **kwargs)
    save_feature_set_links(file)


@property  # type: ignore
@doc_args(File.path.__doc__)
def path(self) -> Union[Path, UPath]:
    """{}"""
    return filepath_from_file(self)


# adapted from: https://stackoverflow.com/questions/9727673
@classmethod  # type: ignore
@doc_args(File.view_tree.__doc__)
def view_tree(
    cls: File,
    path: Optional[PathLike] = None,
    *,
    level: int = -1,
    limit_to_directories: bool = False,
    length_limit: int = 1000,
) -> None:
    """{}"""
    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "
    max_files_per_dir_per_type = 7

    if path is None:
        dir_path = settings.storage
    else:
        dir_path = create_path(path)  # returns Path for local
    n_files = 0
    n_directories = 0

    # by default only including registered files
    # need a flag and a proper implementation
    registered_paths: Set[Any] = set()
    registered_dirs: Set[Any] = set()
    if path is None:
        registered_paths = {
            file.path for file in cls.filter(storage_id=setup_settings.storage.id).all()
        }
        registered_dirs = {d for p in registered_paths for d in p.parents}
    suffixes = set()

    def inner(dir_path: Union[Path, UPath], prefix: str = "", level=-1):
        nonlocal n_files, n_directories, suffixes
        if not level:
            return  # 0, stop iterating
        stripped_dir_path = dir_path.as_posix().rstrip("/")
        # do not iterate through zarr directories
        if stripped_dir_path.endswith((".zarr", ".zrad")):
            return
        # this is needed so that the passed folder is not listed
        contents = [
            i
            for i in dir_path.iterdir()
            if i.as_posix().rstrip("/") != stripped_dir_path
        ]
        if limit_to_directories:
            contents = [d for d in contents if d.is_dir()]
        pointers = [tee] * (len(contents) - 1) + [last]
        n_files_per_dir_per_type = defaultdict(lambda: 0)  # type: ignore
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                if registered_dirs and path not in registered_dirs:
                    continue
                yield prefix + pointer + path.name
                n_directories += 1
                n_files_per_dir_per_type = defaultdict(lambda: 0)
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                if registered_paths and path not in registered_paths:
                    continue
                suffix = extract_suffix_from_path(path)
                suffixes.add(suffix)
                n_files_per_dir_per_type[suffix] += 1
                n_files += 1
                if n_files_per_dir_per_type[suffix] == max_files_per_dir_per_type:
                    yield prefix + "..."
                elif n_files_per_dir_per_type[suffix] > max_files_per_dir_per_type:
                    continue
                else:
                    yield prefix + pointer + path.name

    folder_tree = ""
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        folder_tree += f"\n{line}"
    if next(iterator, None):
        folder_tree += f"... length_limit, {length_limit}, reached, counted:"
    directory_info = "directory" if n_directories == 1 else "directories"
    display_suffixes = ", ".join([f"{suffix!r}" for suffix in suffixes])
    logger.print(
        f"{dir_path.name} ({n_directories} sub-{directory_info} & {n_files} files with"
        f" suffixes {display_suffixes}): {folder_tree}"
    )


METHOD_NAMES = [
    "__init__",
    "from_anndata",
    "from_df",
    "backed",
    "stage",
    "load",
    "delete",
    "save",
    "replace",
    "from_dir",
    "view_tree",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(File, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, File, globals())

# privates currently dealt with separately
File._delete_skip_storage = _delete_skip_storage
File._save_skip_storage = _save_skip_storage
setattr(File, "path", path)



import pytest

import lamindb_setup as ln_setup


def test_delete_invalid_name():
    with pytest.raises(ValueError) as error:
        ln_setup.delete("invalid/name")
    assert (
        error.exconly()
        == "ValueError: Invalid instance name: '/' delimiter not allowed."
    )



import lamindb as ln
from lamindb.dev._data import add_transform_to_kwargs


def test_add_transform_to_kwargs():
    kwargs = {}
    transform = ln.Transform(name="hello")
    transform.save()
    run = ln.Run(transform)
    add_transform_to_kwargs(kwargs, run)
    assert kwargs["transform"] == transform



from functools import cached_property

from lnschema_core import Modality

from .._save import save
from ._settings import settings

MODALITIES = ["meta"]
NUMBER_TYPE = "number"


# might be a bad design idea & disappear again
class Priors:
    @cached_property
    def modalities(self):
        all_defaults = Modality.filter(name__in=MODALITIES).all()
        # update the registry
        if len(all_defaults) != len(MODALITIES):
            upon_create_search_names = settings.upon_create_search_names
            settings.upon_create_search_names = False
            modalities = [Modality(name=name) for name in MODALITIES]
            settings.upon_create_search_names = upon_create_search_names
            save(modalities)
        return Modality.lookup()


priors = Priors()



import pytest

from lamindb.dev.versioning import set_version


def test_set_version():
    # all remaining lines are covered in notebooks
    with pytest.raises(ValueError):
        set_version(None, "1.2")
    assert set_version(None, "0") == "1"
    assert set_version(None, "1") == "2"
    assert set_version("1.2.3", "0") == "1.2.3"
    assert set_version("1.2.3") == "1.2.3"



import os
import pytest
import lamindb_setup as ln_setup
from gotrue.errors import AuthUnknownError
from lamindb_setup import login, settings
from lamindb_setup.dev._hub_core import (
    load_instance,
    sign_in_hub,
)


def test_switch_env():
    # testuser1 is defined in prod & staging with the same password
    assert os.getenv("LAMIN_ENV") == "prod"

    # check whether we can log in
    login("static-testuser1@lamin.ai", password="static-testuser1-password")
    assert settings.user.email == "static-testuser1@lamin.ai"

    # testuser1.staging is defined only in staging
    os.environ["LAMIN_ENV"] = "staging"

    login("testuser1.staging@lamin.ai", password="password")
    assert settings.user.email == "testuser1.staging@lamin.ai"

    # back to prod
    os.environ["LAMIN_ENV"] = "prod"


def test_load_instance_fallbacks():
    prod_url = ln_setup.dev._hub_client.PROD_URL
    ln_setup.dev._hub_client.PROD_URL = (  # deactivated prod url
        "https://inactive.lamin.ai"
    )
    with pytest.raises(AuthUnknownError):
        ln_setup.dev._hub_client.connect_hub_with_auth(renew_token=True)
    assert not isinstance(
        sign_in_hub(
            email="static-testuser1@lamin.ai", password="static-testuser1-password"
        ),
        str,
    )
    load_instance(
        owner="static-testuser1",
        name="static-testinstance1",
    )
    ln_setup.dev._hub_client.PROD_URL = prod_url



import builtins
from typing import List, Optional, Set, Union

from lamin_utils import logger
from lnschema_core import Dataset, File, Registry, Run, Transform
from lnschema_core.models import HasParents, format_field_value

from lamindb._utils import attach_func_to_class_method

from . import _TESTING
from ._registry import StrField, get_default_str_field

LAMIN_GREEN_LIGHTER = "#10b981"
LAMIN_GREEN_DARKER = "#065f46"
GREEN_FILL = "honeydew"
TRANSFORM_EMOJIS = {"notebook": "📔", "app": "🖥️", "pipeline": "🧩"}
is_run_from_ipython = getattr(builtins, "__IPYTHON__", False)


def _transform_emoji(transform: Transform):
    if transform is not None:
        return TRANSFORM_EMOJIS.get(transform.type, "💫")
    else:
        return TRANSFORM_EMOJIS["pipeline"]


def _view(u):
    try:
        if is_run_from_ipython:
            from IPython import get_ipython
            from IPython.display import display

            #  True if the code is running in a Jupyter Notebook or Lab environment
            if get_ipython().__class__.__name__ == "TerminalInteractiveShell":
                return u.view()
            else:
                display(u)
        else:
            return u
    except (FileNotFoundError, RuntimeError):  # pragma: no cover
        logger.error(
            "please install the graphviz executable on your system:\n  - Ubuntu: `sudo"
            " apt-get install graphviz`\n  - Windows:"
            " https://graphviz.org/download/#windows\n  - Mac: `brew install graphviz`"
        )


def view_parents(
    self,
    field: Optional[StrField] = None,
    with_children: bool = False,
    distance: int = 5,
):
    if field is None:
        field = get_default_str_field(self)
    if not isinstance(field, str):
        field = field.field.name

    return _view_parents(
        record=self, field=field, with_children=with_children, distance=distance
    )


def view_flow(data: Union[File, Dataset], with_children: bool = True) -> None:
    """Graph of data flow.

    Notes:
        For more info, see use cases: :doc:`docs:data-flow`.

    Examples:
        >>> dataset.view_flow()
        >>> file.view_flow()
    """
    import graphviz

    df_values = _get_all_parent_runs(data)
    if with_children:
        df_values += _get_all_child_runs(data)
    df_edges = _df_edges_from_runs(df_values)

    data_label = _label_data_run_transform(data)

    u = graphviz.Digraph(
        data.id,
        node_attr={
            "fillcolor": GREEN_FILL,
            "color": LAMIN_GREEN_DARKER,
            "fontname": "Helvetica",
            "fontsize": "10",
        },
        edge_attr={"arrowsize": "0.5"},
    )

    def add_node(
        record: Union[Run, File, Dataset],
        node_id: str,
        node_label: str,
        u: graphviz.Digraph,
    ):
        if isinstance(record, Run):
            fillcolor = "gainsboro"
        else:
            fillcolor = GREEN_FILL
        u.node(
            node_id,
            label=node_label,
            shape="box",
            style="rounded,filled",
            fillcolor=fillcolor,
        )

    for _, row in df_edges.iterrows():
        add_node(row["source_record"], row["source"], row["source_label"], u)
        if row["target_record"] not in df_edges["source_record"]:
            add_node(row["target_record"], row["target"], row["target_label"], u)

        u.edge(row["source"], row["target"], color="dimgrey")
    # label the searched file
    u.node(
        data.id,
        label=data_label,
        style="rounded,filled",
        fillcolor=LAMIN_GREEN_LIGHTER,
        shape="box",
    )

    _view(u)


def _view_parents(
    record: Registry, field: str, with_children: bool = False, distance: int = 100
):
    """Graph of parents."""
    if not hasattr(record, "parents"):
        raise NotImplementedError(
            f"Parents view is not supported for {record.__class__.__name__}!"
        )
    import graphviz
    import pandas as pd

    df_edges = None
    df_edges_parents = _df_edges_from_parents(
        record=record, field=field, distance=distance
    )
    if df_edges_parents is not None:
        df_edges = df_edges_parents
    if with_children:
        df_edges_children = _df_edges_from_parents(
            record=record, field=field, distance=distance, children=True
        )
        if df_edges_children is not None:
            if df_edges is not None:
                df_edges = pd.concat(
                    [df_edges_parents, df_edges_children]
                ).drop_duplicates()
            else:
                df_edges = df_edges_children

    record_label = record.__getattribute__(field)

    u = graphviz.Digraph(
        record.id,
        node_attr={
            "color": LAMIN_GREEN_DARKER,
            "fillcolor": GREEN_FILL,
            "shape": "box",
            "style": "rounded,filled",
            "fontname": "Helvetica",
            "fontsize": "10",
        },
        edge_attr={"arrowsize": "0.5"},
    )
    u.node(
        record.id,
        label=_label_data_run_transform(record)
        if record.__class__.__name__ == "Transform"
        else _add_emoji(record, record_label),
        fillcolor=LAMIN_GREEN_LIGHTER,
    )
    if df_edges is not None:
        for _, row in df_edges.iterrows():
            u.node(row["source"], label=row["source_label"])
            u.node(row["target"], label=row["target_label"])
            u.edge(row["source"], row["target"], color="dimgrey")

    _view(u)


def _get_parents(record: Registry, field: str, distance: int, children: bool = False):
    """Recursively get parent records within a distance."""
    if children:
        key = "parents"
    else:
        key = "children"
    model = record.__class__
    condition = f"{key}__{field}"
    results = model.filter(**{condition: record.__getattribute__(field)}).all()
    if distance < 2:
        return results

    d = 2
    while d < distance:
        condition = f"{key}__{condition}"
        records = model.filter(**{condition: record.__getattribute__(field)}).all()

        if len(records) == 0:
            return results

        results = results | records
        d += 1
    return results


def _df_edges_from_parents(
    record: Registry, field: str, distance: int, children: bool = False
):
    """Construct a DataFrame of edges as the input of graphviz.Digraph."""
    key = "children" if children else "parents"
    parents = _get_parents(
        record=record, field=field, distance=distance, children=children
    )
    all = record.__class__.objects
    records = parents | all.filter(id=record.id)
    df = records.distinct().df(include=[f"{key}__id"])
    if f"{key}__id" not in df.columns:
        return None
    df_edges = df[[f"{key}__id"]]
    df_edges = df_edges.explode(f"{key}__id")
    df_edges.index.name = "target"
    df_edges = df_edges.reset_index()
    df_edges.dropna(axis=0, inplace=True)
    df_edges.rename(columns={f"{key}__id": "source"}, inplace=True)
    df_edges = df_edges.drop_duplicates()

    # colons messes with the node formatting:
    # https://graphviz.readthedocs.io/en/stable/node_ports.html
    df_edges["source_record"] = df_edges["source"].apply(lambda x: all.get(id=x))
    df_edges["target_record"] = df_edges["target"].apply(lambda x: all.get(id=x))
    if record.__class__.__name__ == "Transform":
        df_edges["source_label"] = df_edges["source_record"].apply(
            _label_data_run_transform
        )
        df_edges["target_label"] = df_edges["target_record"].apply(
            _label_data_run_transform
        )
    else:
        df_edges["source_label"] = df_edges["source_record"].apply(
            lambda x: x.__getattribute__(field)
        )
        df_edges["target_label"] = df_edges["target_record"].apply(
            lambda x: x.__getattribute__(field)
        )
    return df_edges


def _add_emoji(record: Registry, label: str):
    if record.__class__.__name__ == "Transform":
        emoji = TRANSFORM_EMOJIS.get(record.type, "💫")
    elif record.__class__.__name__ == "Run":
        emoji = TRANSFORM_EMOJIS.get(record.transform.type, "💫")
    else:
        emoji = ""
    return f"{emoji} {label}"


def _get_all_parent_runs(data: Union[File, Dataset]) -> List:
    """Get all input file/dataset runs recursively."""
    run_inputs_outputs = []

    runs = [data.run] if data.run is not None else []
    while len(runs) > 0:
        inputs = []
        for r in runs:
            inputs_run = r.input_files.list() + r.input_datasets.list()
            run_inputs_outputs += [(inputs_run, r)]
            run_inputs_outputs += [
                (r, r.output_files.list() + r.output_datasets.list())
            ]
            inputs += inputs_run
        runs = [f.run for f in inputs if f.run is not None]
    return run_inputs_outputs


def _get_all_child_runs(data: Union[File, Dataset]) -> List:
    """Get all output file/dataset runs recursively."""
    all_runs: Set[Run] = set()
    run_inputs_outputs = []

    runs = {f.run for f in data.run.output_files.all()}
    runs.update({f.run for f in data.run.output_datasets.all()})
    while runs.difference(all_runs):
        all_runs.update(runs)
        child_runs: Set[Run] = set()
        for r in runs:
            run_inputs_outputs += [(r.input_files.list() + r.input_datasets.list(), r)]
            run_outputs = r.output_files
            run_inputs_outputs += [(r, run_outputs.list())]
            child_runs.update(
                Run.filter(input_files__id__in=r.output_files.list("id")).list()
            )
            child_runs.update(
                Run.filter(input_datasets__id__in=r.output_datasets.list("id")).list()
            )
        runs = child_runs
    return run_inputs_outputs


def _label_data_run_transform(record: Union[File, Run, Transform]):
    if isinstance(record, File):
        if record.description is None:
            name = record.key
        else:
            name = record.description.replace("&", "&amp;")

        return (
            rf'<📄 {name}<BR/><FONT COLOR="GREY" POINT-SIZE="10"'
            rf' FACE="Monospace">id={record.id}<BR/>suffix={record.suffix}</FONT>>'
        )
    elif isinstance(record, Dataset):
        name = record.name.replace("&", "&amp;")
        return (
            rf'<🍱 {name}<BR/><FONT COLOR="GREY" POINT-SIZE="10"'
            rf' FACE="Monospace">id={record.id}<BR/>version={record.version}</FONT>>'
        )
    elif isinstance(record, Run):
        name = f'{record.transform.name.replace("&", "&amp;")}'
        return (
            rf'<{TRANSFORM_EMOJIS.get(str(record.transform.type), "💫")} {name}<BR/><FONT COLOR="GREY" POINT-SIZE="10"'  # noqa
            rf' FACE="Monospace">id={record.id}<BR/>type={record.transform.type},'
            rf" user={record.created_by.name}<BR/>run_at={format_field_value(record.run_at)}</FONT>>"  # noqa
        )
    elif isinstance(record, Transform):
        name = f'{record.name.replace("&", "&amp;")}'
        return (
            rf'<{TRANSFORM_EMOJIS.get(str(record.type), "💫")} {name}<BR/><FONT COLOR="GREY" POINT-SIZE="10"'  # noqa
            rf' FACE="Monospace">id={record.id}<BR/>type={record.type},'
            rf" user={record.created_by.name}<BR/>updated_at={format_field_value(record.updated_at)}</FONT>>"  # noqa
        )


def _df_edges_from_runs(df_values: List):
    import pandas as pd

    df = pd.DataFrame(df_values, columns=["source_record", "target_record"])
    df = df.explode("source_record")
    df = df.explode("target_record")
    df = df.drop_duplicates().dropna()
    df["source"] = [i.id for i in df["source_record"]]
    df["target"] = [i.id for i in df["target_record"]]
    df["source_label"] = df["source_record"].apply(_label_data_run_transform)
    df["target_label"] = df["target_record"].apply(_label_data_run_transform)
    return df


METHOD_NAMES = [
    "view_parents",
]

if _TESTING:  # type: ignore
    from inspect import signature

    SIGS = {
        name: signature(getattr(HasParents, name))
        for name in METHOD_NAMES
        if not name.startswith("__")
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, HasParents, globals())



import os
from uuid import UUID

import pytest
from gotrue.errors import AuthApiError
import lamindb_setup as ln_setup
from lamindb_setup.dev._hub_client import (
    Environment,
    connect_hub_with_auth,
)
from lamindb_setup.dev._hub_core import (
    add_storage,
    init_instance,
    load_instance,
    sign_in_hub,
    sign_up_hub,
)
from lamindb_setup.dev._hub_crud import (
    sb_select_collaborator,
    sb_select_db_user_by_instance,
    sb_select_instance_by_name,
)

# typing
# from lamindb.dev import UserSettings
# from supabase import Client
from lamindb_setup.dev._hub_utils import LaminDsn, base62


def test_runs_locally():
    assert os.environ["LAMIN_ENV"] == "local"
    assert Environment().lamin_env == "local"


def test_incomplete_signup():
    email = "testuser-incomplete-signup@gmail.com"
    response = sign_up_hub(email)
    assert isinstance(response, tuple) and len(response) == 3
    response = sign_in_hub(email, response[0])
    assert response == "complete-signup"


@pytest.fixture(scope="session")
def create_testuser1_session():  # -> Tuple[Client, UserSettings]
    email = "testuser1@gmail.com"
    response = ln_setup.signup(email)
    assert response is None
    account_id = ln_setup.settings.user.uuid.hex
    account = {
        "id": account_id,
        "user_id": account_id,
        "lnid": base62(8),
        "handle": "testuser1",
    }
    # uses ln_setup.settings.user.access_token
    client = connect_hub_with_auth()
    client.table("account").insert(account).execute()
    yield client, ln_setup.settings.user
    client.auth.sign_out()


@pytest.fixture(scope="session")
def create_myinstance(create_testuser1_session):  # -> Dict
    init_instance(
        name="myinstance",
        storage="s3://lndb-setup-ci",
        db="postgresql://postgres:pwd@fakeserver.xyz:5432/mydb",
    )
    client, _ = create_testuser1_session
    instance = sb_select_instance_by_name(
        account_id=ln_setup.settings.user.uuid,
        name="myinstance",
        client=client,
    )
    yield instance


def test_connection_string_decomp(create_myinstance, create_testuser1_session):
    client, _ = create_testuser1_session
    db_user = sb_select_db_user_by_instance(
        instance_id=create_myinstance["id"],
        client=client,
    )
    assert create_myinstance["db_scheme"] == "postgresql"
    assert create_myinstance["db_host"] == "fakeserver.xyz"
    assert create_myinstance["db_port"] == 5432
    assert create_myinstance["db_database"] == "mydb"
    assert db_user["db_user_name"] == "postgres"
    assert db_user["db_user_password"] == "pwd"

    db_collaborator = sb_select_collaborator(
        instance_id=create_myinstance["id"],
        account_id=ln_setup.settings.user.uuid.hex,
        client=client,
    )
    assert db_collaborator["db_user_id"] == db_user["id"]


def test_load_instance(create_myinstance, create_testuser1_session):
    # trigger return for inexistent handle
    assert "account-not-exists" == load_instance(
        owner="testusr1",  # testuser1 with a typo
        name=create_myinstance["name"],
    )
    # trigger misspelled name
    assert "instance-not-reachable" == load_instance(
        owner="testuser1",
        name="inexistent-name",  # inexistent name
    )
    # now supply correct data
    result = load_instance(
        owner="testuser1",
        name=create_myinstance["name"],
    )
    client, _ = create_testuser1_session
    db_user = sb_select_db_user_by_instance(
        instance_id=create_myinstance["id"],
        client=client,
    )
    expected_dsn = LaminDsn.build(
        scheme=create_myinstance["db_scheme"],
        user=db_user["db_user_name"],
        password=db_user["db_user_password"],
        host=create_myinstance["db_host"],
        port=str(create_myinstance["db_port"]),
        database=create_myinstance["db_database"],
    )
    loaded_instance, _ = result
    assert loaded_instance["name"] == create_myinstance["name"]
    assert loaded_instance["db"] == expected_dsn


def test_load_instance_corrupted_or_expired_credentials(
    create_myinstance, create_testuser1_session
):
    # assume token & password are corrupted or expired
    ln_setup.settings.user.access_token = "corrupted_or_expired_token"
    correct_password = ln_setup.settings.user.password
    ln_setup.settings.user.password = "corrupted_password"
    with pytest.raises(AuthApiError):
        load_instance(
            owner="testuser1",
            name=create_myinstance["name"],
        )
    # now, let's assume only the token is expired or corrupted
    # re-creating the auth client triggers a re-generated token because it
    # excepts the error assuming the token is expired
    ln_setup.settings.user.access_token = "corrupted_or_expired_token"
    ln_setup.settings.user.password = correct_password
    load_instance(
        owner="testuser1",
        name=create_myinstance["name"],
    )


def test_add_storage(create_testuser1_session):
    client, usettings = create_testuser1_session
    storage_id = add_storage(
        root="s3://lndb-setup-ci",
        account_id=usettings.uuid,
        hub=client,
    )
    assert isinstance(storage_id, UUID)


def test_add_storage_with_non_existing_bucket(create_testuser1_session):
    client, usettings = create_testuser1_session
    from botocore.exceptions import ClientError

    with pytest.raises(ClientError) as error:
        add_storage(
            root="s3://non_existing_storage_root",
            account_id=usettings.uuid,
            hub=client,
        )
    assert error.exconly().endswith("Not Found")



import os
import shutil
from pathlib import Path
from typing import Literal, Optional, Set, Tuple, Union
from uuid import UUID
from ._hub_client import call_with_fallback
from ._hub_crud import sb_select_account_handle_name_by_lnid
from lamin_utils import logger
from .cloud_sqlite_locker import (
    InstanceLockedException,
    EXPIRATION_TIME,
)
from ._settings_save import save_instance_settings
from ._settings_storage import StorageSettings
from ._settings_store import current_instance_settings_file, instance_settings_file
from .upath import UPath


class InstanceSettings:
    """Instance settings."""

    def __init__(
        self,
        owner: str,  # owner handle
        name: str,  # instance name
        storage_root: Union[str, Path, UPath],  # storage location on cloud
        storage_region: Optional[str] = None,
        db: Optional[str] = None,  # DB URI
        schema: Optional[str] = None,  # comma-separated string of schema names
        id: Optional[UUID] = None,  # instance id
    ):
        self._owner: str = owner
        self._name: str = name
        self._storage: StorageSettings = StorageSettings(
            storage_root, region=storage_region
        )
        self._db: Optional[str] = db
        self._schema_str: Optional[str] = schema
        self._id: Optional[UUID] = id

    def __repr__(self):
        """Rich string representation."""
        representation = f"Current instance: {self.identifier}"
        attrs = ["owner", "name", "storage", "db", "schema"]
        for attr in attrs:
            value = getattr(self, attr)
            if attr == "storage":
                representation += f"\n- storage root: {value.root_as_str}"
                representation += f"\n- storage region: {value.region}"
            else:
                representation += f"\n- {attr}: {value}"
        return representation

    @property
    def owner(self) -> str:
        """Instance owner. A user or organization account handle."""
        return self._owner

    @property
    def name(self) -> str:
        """Instance name."""
        return self._name

    @property
    def identifier(self) -> str:
        """Unique semantic identifier.

        See remote instances at https://lamin.ai/owner/name.
        """
        return f"{self.owner}/{self.name}"

    @property
    def id(self) -> Optional[UUID]:
        """The instance id."""
        return self._id

    @property
    def schema(self) -> Set[str]:
        """Schema modules in addition to core schema."""
        if self._schema_str is None:
            return {}  # type: ignore
        else:
            return {schema for schema in self._schema_str.split(",") if schema != ""}

    @property
    def _sqlite_file(self) -> UPath:
        """SQLite file."""
        return self.storage.key_to_filepath(f"{self.name}.lndb")

    @property
    def _sqlite_file_local(self) -> Path:
        """Local SQLite file."""
        return self.storage.cloud_to_local_no_update(self._sqlite_file)

    def _update_cloud_sqlite_file(self, unlock_cloud_sqlite: bool = True) -> None:
        """Upload the local sqlite file to the cloud file."""
        if self._is_cloud_sqlite:
            sqlite_file = self._sqlite_file
            logger.warning(
                f"updating & unlocking cloud SQLite '{sqlite_file}' of instance"
                f" '{self.identifier}'"
            )
            cache_file = self.storage.cloud_to_local_no_update(sqlite_file)
            sqlite_file.upload_from(cache_file)  # type: ignore
            cloud_mtime = sqlite_file.modified.timestamp()  # type: ignore
            # this seems to work even if there is an open connection
            # to the cache file
            os.utime(cache_file, times=(cloud_mtime, cloud_mtime))
            if unlock_cloud_sqlite:
                self._cloud_sqlite_locker.unlock()

    def _update_local_sqlite_file(self, lock_cloud_sqlite: bool = True) -> None:
        """Download the cloud sqlite file if it is newer than local."""
        if self._is_cloud_sqlite:
            logger.warning(
                "updating local SQLite & locking cloud SQLite (sync back & unlock:"
                " lamin close)"
            )
            if lock_cloud_sqlite:
                self._cloud_sqlite_locker.lock()
                self._check_sqlite_lock()
            sqlite_file = self._sqlite_file
            cache_file = self.storage.cloud_to_local_no_update(sqlite_file)
            sqlite_file.synchronize(cache_file)  # type: ignore

    def _check_sqlite_lock(self):
        if not self._cloud_sqlite_locker.has_lock:
            locked_by = self._cloud_sqlite_locker._locked_by
            lock_msg = "Cannot load the instance, it is locked by "
            user_info = call_with_fallback(
                sb_select_account_handle_name_by_lnid,
                lnid=locked_by,
            )
            if user_info is None:
                lock_msg += f"id: '{locked_by}'."
            else:
                lock_msg += (
                    f"'{user_info['handle']}' (id: '{locked_by}', name:"
                    f" '{user_info['name']}')."
                )
            lock_msg += (
                "The instance will be automatically unlocked after"
                f" {int(EXPIRATION_TIME/3600/24)}d of no activity."
            )
            raise InstanceLockedException(lock_msg)

    @property
    def db(self) -> str:
        """Database connection string (URI)."""
        if self._db is None:
            # here, we want the updated sqlite file
            # hence, we don't use self._sqlite_file_local()
            return f"sqlite:///{self.storage.cloud_to_local(self._sqlite_file)}"
        else:
            return self._db

    @property
    def dialect(self) -> Literal["sqlite", "postgresql"]:
        """SQL dialect."""
        if self._db is None or self._db.startswith("sqlite://"):
            return "sqlite"
        else:
            assert self._db.startswith("postgresql")
            return "postgresql"

    @property
    def session(self):
        raise NotImplementedError

    @property
    def _is_cloud_sqlite(self) -> bool:
        # can we make this a private property, Sergei?
        # as it's not relevant to the user
        """Is this a cloud instance with sqlite db."""
        return self.dialect == "sqlite" and self.storage.is_cloud

    @property
    def _cloud_sqlite_locker(self):
        # avoid circular import
        from .cloud_sqlite_locker import empty_locker, get_locker

        if self._is_cloud_sqlite:
            try:
                return get_locker(self)
            except PermissionError:
                logger.warning("read-only access - did not access locker")
                return empty_locker
        else:
            return empty_locker

    @property
    def storage(self) -> StorageSettings:
        """Low-level access to storage location."""
        return self._storage

    @property
    def is_remote(self) -> bool:
        """Boolean indicating if an instance has no local component."""
        if not self.storage.is_cloud:
            return False

        def is_local_uri(uri: str):
            if "@localhost:" in uri:
                return True
            if "@0.0.0.0:" in uri:
                return True
            if "@127.0.0.1" in uri:
                return True
            return False

        if self.dialect == "postgresql":
            if is_local_uri(self.db):
                return False
        # returns True for cloud SQLite
        # and remote postgres
        return True

    def _get_settings_file(self) -> Path:
        return instance_settings_file(self.name, self.owner)

    def _persist(self) -> None:
        assert self.name is not None
        if self.storage.type == "local":
            self.storage.root.mkdir(parents=True, exist_ok=True)

        filepath = self._get_settings_file()
        # persist under filepath for later reference
        save_instance_settings(self, filepath)
        # persist under current file for auto load
        shutil.copy2(filepath, current_instance_settings_file())
        # persist under settings class for same session reference
        # need to import here to avoid circular import
        from .._settings import settings

        settings._instance_settings = self

    def _init_db(self):
        from .django import setup_django

        setup_django(self, deploy_migrations=True, init=True)

    def _load_db(
        self, do_not_lock_for_laminapp_admin: bool = False
    ) -> Tuple[bool, str]:
        # Is the database available and initialized as LaminDB?
        # returns a tuple of status code and message
        if self.dialect == "sqlite" and not self._sqlite_file.exists():
            return False, "SQLite file does not exist"
        from lamindb_setup import settings  # to check user
        from .django import setup_django

        # lock in all cases except if do_not_lock_for_laminapp_admin is True and
        # user is `laminapp-admin`
        # value doesn't matter if not a cloud sqlite instance
        lock_cloud_sqlite = self._is_cloud_sqlite and (
            not do_not_lock_for_laminapp_admin
            or settings.user.handle != "laminapp-admin"
        )
        # we need the local sqlite to setup django
        self._update_local_sqlite_file(lock_cloud_sqlite=lock_cloud_sqlite)
        # setting up django also performs a check for migrations & prints them
        # as warnings
        # this should fail, e.g., if the db is not reachable
        setup_django(self)
        return True, ""



from typing import Dict, Iterable, List, Literal, Optional, Set, Union

import numpy as np
import pandas as pd
from django.core.exceptions import FieldDoesNotExist
from django.db.models import QuerySet
from lamin_utils import colors, logger
from lamin_utils._inspect import InspectResult
from lamindb_setup.dev._docs import doc_args
from lnschema_core import CanValidate, Registry
from lnschema_core.types import ListLike, StrField

from lamindb._utils import attach_func_to_class_method

from . import _TESTING
from ._from_values import _has_species_field, _print_values
from ._registry import _queryset, get_default_str_field


@classmethod  # type: ignore
@doc_args(CanValidate.inspect.__doc__)
def inspect(
    cls,
    values: ListLike,
    field: Optional[Union[str, StrField]] = None,
    *,
    mute: bool = False,
    **kwargs,
) -> InspectResult:
    """{}"""
    return _inspect(
        cls=cls,
        values=values,
        field=field,
        mute=mute,
        **kwargs,
    )


@classmethod  # type: ignore
@doc_args(CanValidate.validate.__doc__)
def validate(
    cls,
    values: ListLike,
    field: Optional[Union[str, StrField]] = None,
    *,
    mute: bool = False,
    **kwargs,
) -> np.ndarray:
    """{}"""
    return _validate(cls=cls, values=values, field=field, mute=mute, **kwargs)


def _inspect(
    cls,
    values: ListLike,
    field: Optional[Union[str, StrField]] = None,
    *,
    mute: bool = False,
    **kwargs,
) -> Union["pd.DataFrame", Dict[str, List[str]]]:
    """{}"""
    from lamin_utils._inspect import inspect

    if isinstance(values, str):
        values = [values]

    field = get_default_str_field(cls, field=field)
    queryset = _queryset(cls)
    orm = queryset.model
    model_name = orm._meta.model.__name__

    # inspect in the DB
    result_db = inspect(
        df=_filter_query_based_on_species(
            queryset=queryset, species=kwargs.get("species")
        ),
        identifiers=values,
        field=field,
        mute=mute,
        **kwargs,
    )
    nonval = set(result_db.non_validated).difference(result_db.synonyms_mapper.keys())

    if len(nonval) > 0 and orm.__get_schema_name__() == "bionty":
        bionty_result = orm.bionty(species=kwargs.get("species")).inspect(
            values=nonval, field=field, mute=True, **kwargs
        )
        bionty_validated = bionty_result.validated
        bionty_mapper = bionty_result.synonyms_mapper
        hint = False
        if len(bionty_validated) > 0 and not mute:
            print_values = _print_values(bionty_validated)
            s = "" if len(bionty_validated) == 1 else "s"
            labels = colors.yellow(f"{len(bionty_validated)} {model_name} term{s}")
            logger.print(
                f"   detected {labels} in Bionty for"
                f" {colors.italic(field)}: {colors.yellow(print_values)}"
            )
            hint = True

        if len(bionty_mapper) > 0 and not mute:
            print_values = _print_values(list(bionty_mapper.keys()))
            s = "" if len(bionty_mapper) == 1 else "s"
            labels = colors.yellow(f"{len(bionty_mapper)} {model_name} term{s}")
            logger.print(
                f"   detected {labels} in Bionty as {colors.italic(f'synonym{s}')}:"
                f" {colors.yellow(print_values)}"
            )
            hint = True

        if hint:
            logger.print(
                f"→  add records from Bionty to your {model_name} registry via"
                f" {colors.italic('.from_values()')}"
            )

        nonval = bionty_result.non_validated

    if len(nonval) > 0 and not mute:
        print_values = _print_values(list(nonval))
        s = "" if len(nonval) == 1 else "s"
        labels = colors.red(f"{len(nonval)} term{s}")
        logger.print(f"   couldn't validate {labels}: {colors.red(print_values)}")
        logger.print(
            f"→  if you are sure, create new record{s} via"
            f" {colors.italic(f'ln.{orm.__name__}()')} and save to your registry"
        )

    return result_db


def _validate(
    cls,
    values: ListLike,
    field: Optional[Union[str, StrField]] = None,
    *,
    mute: bool = False,
    **kwargs,
) -> np.ndarray:
    """{}"""
    from lamin_utils._inspect import validate

    return_str = True if isinstance(values, str) else False
    if isinstance(values, str):
        values = [values]

    field = get_default_str_field(cls, field=field)

    queryset = _queryset(cls)
    field_values = pd.Series(
        _filter_query_based_on_species(
            queryset=queryset,
            species=kwargs.get("species"),
            values_list_field=field,
        ),
        dtype="object",
    )

    result = validate(
        identifiers=values,
        field_values=field_values,
        case_sensitive=True,
        mute=mute,
        field=field,
        **kwargs,
    )
    if return_str and len(result) == 1:
        return result[0]
    else:
        return result


@classmethod  # type: ignore
@doc_args(CanValidate.standardize.__doc__)
def standardize(
    cls,
    values: Iterable,
    field: Optional[Union[str, StrField]] = None,
    *,
    return_mapper: bool = False,
    case_sensitive: bool = False,
    mute: bool = False,
    bionty_aware: bool = True,
    keep: Literal["first", "last", False] = "first",
    synonyms_field: str = "synonyms",
    **kwargs,
) -> Union[List[str], Dict[str, str]]:
    """{}"""
    return _standardize(
        cls=cls,
        values=values,
        field=field,
        return_mapper=return_mapper,
        case_sensitive=case_sensitive,
        mute=mute,
        bionty_aware=bionty_aware,
        keep=keep,
        synonyms_field=synonyms_field,
        **kwargs,
    )


def set_abbr(self, value: str):
    self.abbr = value

    if hasattr(self, "name") and value == self.name:
        pass
    else:
        try:
            self.add_synonym(value, save=False)
        except Exception:  # pragma: no cover
            pass

    if not self._state.adding:
        self.save()


def add_synonym(
    self,
    synonym: Union[str, ListLike],
    force: bool = False,
    save: Optional[bool] = None,
):
    _check_synonyms_field_exist(self)
    _add_or_remove_synonyms(
        synonym=synonym, record=self, force=force, action="add", save=save
    )


def remove_synonym(self, synonym: Union[str, ListLike]):
    _check_synonyms_field_exist(self)
    _add_or_remove_synonyms(synonym=synonym, record=self, action="remove")


def _standardize(
    cls,
    values: Iterable,
    field: Optional[Union[str, StrField]] = None,
    *,
    return_mapper: bool = False,
    case_sensitive: bool = False,
    mute: bool = False,
    bionty_aware: bool = True,
    keep: Literal["first", "last", False] = "first",
    synonyms_field: str = "synonyms",
    **kwargs,
) -> Union[List[str], Dict[str, str]]:
    """{}"""
    from lamin_utils._map_synonyms import map_synonyms

    return_str = True if isinstance(values, str) else False
    if isinstance(values, str):
        values = [values]

    field = get_default_str_field(cls, field=field)
    queryset = _queryset(cls)
    orm = queryset.model

    species = kwargs.get("species")
    if _has_species_field(orm):
        # here, we can safely import lnschema_bionty
        from lnschema_bionty._bionty import create_or_get_species_record

        species_record = create_or_get_species_record(species=species, orm=orm)
        species = species_record.name if species_record is not None else species_record

    try:
        orm._meta.get_field(synonyms_field)
        df = _filter_query_based_on_species(queryset=queryset, species=species)
    except FieldDoesNotExist:
        df = pd.DataFrame()

    _kwargs = dict(
        field=field,
        case_sensitive=case_sensitive,
        keep=keep,
        synonyms_field=synonyms_field,
    )
    # standardized names from the DB
    std_names_db = map_synonyms(
        df=df,
        identifiers=values,
        return_mapper=return_mapper,
        mute=mute,
        **_kwargs,
    )

    def _return(result: List, mapper: Dict):
        if return_mapper:
            return mapper
        else:
            if return_str and len(result) == 1:
                return result[0]
            return result

    # map synonyms in Bionty
    if orm.__get_schema_name__() == "bionty" and bionty_aware:
        mapper = {}
        if return_mapper:
            mapper = std_names_db
            std_names_db = map_synonyms(
                df=df, identifiers=values, return_mapper=False, mute=True, **_kwargs
            )

        val_res = orm.validate(std_names_db, field=field, mute=True, species=species)
        if all(val_res):
            return _return(result=std_names_db, mapper=mapper)

        nonval = np.array(std_names_db)[~val_res]
        std_names_bt_mapper = orm.bionty(species=species).standardize(
            nonval, return_mapper=True, mute=True, **_kwargs
        )

        if len(std_names_bt_mapper) > 0 and not mute:
            s = "" if len(std_names_bt_mapper) == 1 else "s"
            warn_msg = (
                f"found {len(std_names_bt_mapper)} synonym{s} in Bionty:"
                f" {list(std_names_bt_mapper.keys())}"
            )
            warn_msg += (
                f"\n   please add corresponding {orm._meta.model.__name__} records via"
                f" `.from_values({list(std_names_bt_mapper.values())})`"
            )
            logger.warning(warn_msg)

        mapper.update(std_names_bt_mapper)
        result = pd.Series(std_names_db).replace(std_names_bt_mapper).tolist()
        return _return(result=result, mapper=mapper)

    else:
        return _return(result=std_names_db, mapper=std_names_db)


def _add_or_remove_synonyms(
    synonym: Union[str, Iterable],
    record: Registry,
    action: Literal["add", "remove"],
    force: bool = False,
    save: Optional[bool] = None,
):
    """Add or remove synonyms."""

    def check_synonyms_in_all_records(synonyms: Set[str], record: Registry):
        """Errors if input synonym is associated with other records in the DB."""
        import pandas as pd
        from IPython.display import display

        syns_all = (
            record.__class__.objects.exclude(synonyms="").exclude(synonyms=None).all()
        )
        if len(syns_all) == 0:
            return
        df = pd.DataFrame(syns_all.values())
        df["synonyms"] = df["synonyms"].str.split("|")
        df = df.explode("synonyms")
        matches_df = df[(df["synonyms"].isin(synonyms)) & (df["id"] != record.id)]
        if matches_df.shape[0] > 0:
            records_df = pd.DataFrame(syns_all.filter(id__in=matches_df["id"]).values())
            logger.error(
                f"input synonyms {matches_df['synonyms'].unique()} already associated"
                " with the following records:\n"
            )
            display(records_df)
            raise SystemExit(AssertionError)

    # passed synonyms
    # nothing happens when passing an empty string or list
    if isinstance(synonym, str):
        if len(synonym) == 0:
            return
        syn_new_set = set([synonym])
    else:
        if synonym == [""]:
            return
        syn_new_set = set(synonym)
    # nothing happens when passing an empty string or list
    if len(syn_new_set) == 0:
        return
    # because we use | as the separator
    if any(["|" in i for i in syn_new_set]):
        raise AssertionError("a synonym can't contain '|'!")

    # existing synonyms
    syns_exist = record.synonyms
    if syns_exist is None or len(syns_exist) == 0:
        syns_exist_set = set()
    else:
        syns_exist_set = set(syns_exist.split("|"))

    if action == "add":
        if not force:
            check_synonyms_in_all_records(syn_new_set, record)
        syns_exist_set.update(syn_new_set)
    elif action == "remove":
        syns_exist_set = syns_exist_set.difference(syn_new_set)

    if len(syns_exist_set) == 0:
        syns_str = None
    else:
        syns_str = "|".join(syns_exist_set)

    record.synonyms = syns_str

    if save is None:
        # if record is already in DB, save the changes to DB
        save = not record._state.adding
    if save:
        record.save()


def _check_synonyms_field_exist(record: Registry):
    try:
        record.__getattribute__("synonyms")
    except AttributeError:
        raise NotImplementedError(
            f"No synonyms field found in table {record.__class__.__name__}!"
        )


def _filter_query_based_on_species(
    queryset: QuerySet,
    species: Optional[Union[str, Registry]] = None,
    values_list_field: Optional[str] = None,
):
    import pandas as pd

    orm = queryset.model

    if _has_species_field(orm):
        # here, we can safely import lnschema_bionty
        from lnschema_bionty._bionty import create_or_get_species_record

        species_record = create_or_get_species_record(species=species, orm=orm)
        if species_record is not None:
            queryset = queryset.filter(species__name=species_record.name)

    if values_list_field is None:
        return pd.DataFrame.from_records(queryset.values())
    else:
        return queryset.values_list(values_list_field, flat=True)


METHOD_NAMES = [
    "validate",
    "inspect",
    "standardize",
    "add_synonym",
    "remove_synonym",
    "set_abbr",
]

if _TESTING:  # type: ignore
    from inspect import signature

    SIGS = {
        name: signature(getattr(CanValidate, name))
        for name in METHOD_NAMES
        if not name.startswith("__")
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, CanValidate, globals())



from typing import Type

from lnschema_core import Registry

from lamindb._query_set import QuerySet


def filter(Registry: Type[Registry], **expressions) -> QuerySet:
    """See :meth:`~lamindb.dev.Registry.filter`."""
    qs = QuerySet(model=Registry)
    if len(expressions) > 0:
        return qs.filter(**expressions)
    else:
        return qs



from inspect import signature

import lnschema_bionty as lb
import pandas as pd
import pytest

import lamindb as ln
from lamindb import _feature_set
from lamindb._feature_set import get_related_name, validate_features

df = pd.DataFrame(
    {
        "feat1": [1, 2, 3],
        "feat2": [3, 4, 5],
        "feat3": ["cond1", "cond2", "cond2"],
        "feat4": ["id1", "id2", "id3"],
    }
)


def test_signatures():
    # this seems currently the easiest and most transparent
    # way to test violations of the signature equality
    # the MockORM class is needed to get inspect.signature
    # to work
    class Mock:
        pass

    # class methods
    class_methods = ["from_values", "from_df"]
    for name in class_methods:
        setattr(Mock, name, getattr(_feature_set, name))
        assert signature(getattr(Mock, name)) == _feature_set.SIGS.pop(name)
    # methods
    for name, sig in _feature_set.SIGS.items():
        assert signature(getattr(_feature_set, name)) == sig


def test_feature_set_from_values():
    gene_symbols = ["TCF7", "MYC"]
    lb.settings.species = "human"
    lb.Gene.filter(symbol__in=gene_symbols).all().delete()
    feature_set = ln.FeatureSet.from_values(gene_symbols, lb.Gene.symbol, type=int)
    assert feature_set is None
    ln.save(lb.Gene.from_values(gene_symbols, "symbol"))
    feature_set = ln.FeatureSet.from_values(gene_symbols, lb.Gene.symbol)
    assert feature_set.type == "number"  # this is NUMBER_TYPE
    feature_set = ln.FeatureSet.from_values(gene_symbols, lb.Gene.symbol, type=int)
    assert feature_set._state.adding
    assert feature_set.type == "number"
    assert feature_set.registry == "bionty.Gene"
    feature_set.save()
    id = feature_set.id
    # test that the feature_set is retrieved from the database
    # in case it already exists
    feature_set = ln.FeatureSet.from_values(gene_symbols, lb.Gene.symbol, type=int)
    assert not feature_set._state.adding
    assert id == feature_set.id
    feature_set.delete()

    # edge cases
    with pytest.raises(ValueError):
        feature_set = ln.FeatureSet.from_values([])

    with pytest.raises(TypeError):
        ln.FeatureSet.from_values(["a"], field="name")
    feature_set = ln.FeatureSet.from_values(
        ["weird_name"], field=ln.Feature.name, type="float"
    )
    assert feature_set is None
    with pytest.raises(TypeError):
        ln.FeatureSet.from_values([1], field=ln.ULabel.name, type="float")

    # return none if no validated features
    assert (
        ln.FeatureSet.from_values(["name"], field=ln.ULabel.name, type="float") is None
    )


def test_feature_set_from_records():
    features = ln.Feature.from_df(df)
    with pytest.raises(ValueError) as error:
        feature_set = ln.FeatureSet(features)
    assert (
        error.exconly()
        == "ValueError: Can only construct feature sets from validated features"
    )

    ln.save(features)
    feature_set = ln.FeatureSet(features)
    id = feature_set.id
    assert feature_set._state.adding
    assert feature_set.type is None
    assert feature_set.registry == "core.Feature"
    feature_set.save()
    # test that the feature_set is retrieved from the database
    # in case it already exists
    feature_set = ln.FeatureSet(features)
    assert not feature_set._state.adding
    assert id == feature_set.id
    feature_set.delete()

    # edge case
    with pytest.raises(ValueError):
        positional_arg = 1
        ln.FeatureSet(features, positional_arg)


def test_feature_set_from_df():
    # test using type
    lb.settings.species = "human"
    genes = [lb.Gene(symbol=name) for name in df.columns]
    ln.save(genes)
    with pytest.raises(ValueError) as error:
        ln.FeatureSet.from_df(df, field=lb.Gene.symbol)
    assert error.exconly().startswith("ValueError: data types are heterogeneous:")
    feature_set = ln.FeatureSet.from_df(df[["feat1", "feat2"]], field=lb.Gene.symbol)
    # clean up
    feature_set.delete()
    for gene in genes:
        gene.delete()

    # now for the features registry
    features = ln.Feature.from_df(df)
    ln.save(features)
    feature_set = ln.FeatureSet.from_df(df)
    feature_set.save()
    assert feature_set.type is None
    for feature in feature_set.features.all():
        feature.delete()
    feature_set.delete()


def test_get_related_name():
    with pytest.raises(ValueError):
        get_related_name(ln.Transform)


def test_validate_features():
    with pytest.raises(ValueError):
        validate_features([])
    with pytest.raises(TypeError):
        validate_features(["feature"])
    with pytest.raises(TypeError):
        validate_features({"feature"})
    transform = ln.Transform(name="test")
    transform.save()
    # This is just a type check
    with pytest.raises(TypeError) as error:
        validate_features([transform, ln.Run(transform)])
    assert error.exconly() == "TypeError: feature_set can only contain a single type"
    transform.delete()


def test_kwargs():
    with pytest.raises(ValueError):
        ln.FeatureSet(x="1", features=[])


def test_edge_cases():
    feature = ln.Feature(name="rna", type="float")
    ln.save([feature])
    # wrong type for modality
    with pytest.raises(TypeError):
        ln.FeatureSet([feature], modality="random")
    # unsaved modality
    result = ln.Modality.filter(name="rna").one_or_none()
    if result is not None:
        result.delete()
    modality = ln.Modality(name="rna")
    with pytest.raises(ValueError) as error:
        ln.FeatureSet([feature], modality=modality)
    assert error.exconly().startswith(
        "ValueError: unvalidated modality, save to registry if you're sure it is"
        " correct: "
    )
    modality.save()
    feature_set = ln.FeatureSet([feature], modality=modality)
    assert feature_set.modality == modality
    with pytest.raises(ValueError) as error:
        feature_set = ln.FeatureSet(feature)
    assert (
        error.exconly()
        == "ValueError: Please pass a ListLike of features, not a single feature"
    )
    feature.delete()
    feature_set.delete()



import importlib
import sys
from pathlib import Path
from typing import Optional, Union
from uuid import UUID

from lamin_utils import logger
from pydantic import PostgresDsn

from lamindb_setup.dev.upath import LocalPathClasses, UPath
from ._close import close as close_instance
from ._docstrings import instance_description as description
from ._settings import settings
from ._silence_loggers import silence_loggers
from .dev import InstanceSettings
from .dev._docs import doc_args
from .dev._settings_storage import StorageSettings
from .dev.upath import create_path


def get_schema_module_name(schema_name) -> str:
    return f"lnschema_{schema_name.replace('-', '_')}"


def register_storage(ssettings: StorageSettings):
    from lnschema_core.models import Storage

    storage, created = Storage.objects.update_or_create(
        root=ssettings.root_as_str,
        defaults=dict(
            root=ssettings.root_as_str,
            type=ssettings.type,
            region=ssettings.region,
            created_by_id=settings.user.id,
        ),
    )
    if created:
        logger.save(f"saved: {storage}")


def register_user(usettings):
    from lnschema_core.models import User

    if usettings.handle != "laminapp-admin":
        user, created = User.objects.update_or_create(
            id=usettings.id,
            defaults=dict(
                handle=usettings.handle,
                name=usettings.name,
                email=usettings.email,
            ),
        )
        if created:
            logger.save(f"saved: {user}")


def register_user_and_storage(isettings: InstanceSettings, usettings):
    """Register user & storage in DB."""
    from django.db.utils import OperationalError

    try:
        register_user(usettings)
        register_storage(isettings.storage)
    except OperationalError as error:
        logger.warning(f"instance seems not set up ({error})")


def reload_schema_modules(isettings: InstanceSettings):
    schema_names = ["core"] + list(isettings.schema)
    schema_module_names = [get_schema_module_name(n) for n in schema_names]

    for schema_module_name in schema_module_names:
        if schema_module_name in sys.modules:
            schema_module = importlib.import_module(schema_module_name)
            importlib.reload(schema_module)


def reload_lamindb(isettings: InstanceSettings):
    # only touch lamindb if we're operating from lamindb
    reload_schema_modules(isettings)
    if "lamindb" in sys.modules:
        import lamindb

        importlib.reload(lamindb)
        logger.important(
            f"loaded instance: {isettings.identifier} (lamindb {lamindb.__version__})"
        )
    else:
        # only log if we're outside lamindb
        # lamindb itself logs upon import!
        logger.important(f"loaded instance: {isettings.owner}/{isettings.name}")


ERROR_SQLITE_CACHE = """
Your cached local SQLite file exists, while your cloud SQLite file ({}) doesn't.
Either delete your cache ({}) or add it back to the cloud (if delete was accidental).
"""


@doc_args(
    description.storage_root,
    description.name,
    description.db,
    description.schema,
)
def init(
    *,
    storage: Union[str, Path, UPath],
    name: Optional[str] = None,
    db: Optional[PostgresDsn] = None,
    schema: Optional[str] = None,
    _test: bool = False,
) -> None:
    """Creating and loading a LaminDB instance.

    Args:
        storage: {}
        name: {}
        db: {}
        schema: {}
    """
    from ._check_instance_setup import check_instance_setup

    if check_instance_setup() and not _test:
        raise RuntimeError(
            "Currently don't support init or load of multiple instances in the same"
            " Python session. We will bring this feature back at some point."
        )
    else:
        close_instance(mute=True)
    # clean up in next refactor
    # avoid circular import
    from ._load_instance import load
    from .dev._hub_core import init_instance as init_instance_hub
    from .dev._hub_utils import (
        get_storage_region,
        validate_db_arg,
        validate_schema_arg,
        validate_storage_root_arg,
    )

    #
    if name is not None and "/" in name:
        logger.warning(
            "Please provide a valid instance name: '/' delimiter not allowed."
        )
        raise ValueError("Invalid instance name: '/' delimiter not allowed.")

    assert settings.user.id  # check user is logged in
    owner = settings.user.handle

    schema = validate_schema_arg(schema)
    validate_storage_root_arg(str(storage))
    validate_db_arg(db)

    name_str = infer_instance_name(storage=storage, name=name, db=db)
    # test whether instance exists by trying to load it
    response = load(
        f"{owner}/{name_str}", _raise_not_reachable_error=False, _test=_test
    )
    if response is None:
        return None  # successful load!

    isettings = InstanceSettings(
        owner=owner,
        name=name_str,
        storage_root=storage,
        storage_region=get_storage_region(storage),  # type: ignore
        db=db,
        schema=schema,
    )

    if isettings._is_cloud_sqlite:
        if (
            not isettings._sqlite_file.exists()
            and isettings._sqlite_file_local.exists()
        ):
            raise RuntimeError(
                ERROR_SQLITE_CACHE.format(
                    isettings._sqlite_file, isettings._sqlite_file_local
                )
            )

    if isettings.is_remote:
        result = init_instance_hub(
            name=name_str,
            storage=str(storage),
            db=db,
            schema=schema,
        )
        if not isinstance(result, UUID):
            raise RuntimeError(f"Registering instance on hub failed:\n{result}")
        isettings._id = result
        logger.save(f"registered instance on hub: https://lamin.ai/{owner}/{name_str}")

    if _test:
        isettings._persist()
        return None

    silence_loggers()

    isettings._init_db()
    load_from_isettings(isettings, init=True)
    if isettings._is_cloud_sqlite:
        isettings._cloud_sqlite_locker.lock()
        logger.warning(
            "locked instance (to unlock and push changes to the cloud SQLite file,"
            " call: lamin close)"
        )
    if not isettings.is_remote:
        verbosity = logger._verbosity
        logger.set_verbosity(4)
        logger.hint(
            "did not register local instance on hub (if you want, call `lamin"
            " register`)"
        )
        logger.set_verbosity(verbosity)
    return None


def load_from_isettings(
    isettings: InstanceSettings,
    *,
    init: bool = False,
) -> None:
    from .dev._setup_bionty_sources import load_bionty_sources, write_bionty_sources

    if init:
        # during init both user and storage need to be registered
        register_user_and_storage(isettings, settings.user)
        write_bionty_sources(isettings)
        isettings._update_cloud_sqlite_file(unlock_cloud_sqlite=False)
    else:
        # when loading, django is already set up
        # only register user if the instance is loaded
        # for the first time in an environment
        # this is our best proxy for that the user might not be
        # yet be registered
        if not isettings._get_settings_file().exists():
            register_user(settings.user)
        load_bionty_sources(isettings)
    isettings._persist()
    reload_lamindb(isettings)


def infer_instance_name(
    *,
    storage: Union[str, Path, UPath],
    name: Optional[str] = None,
    db: Optional[PostgresDsn] = None,
):
    if name is not None:
        return name
    if db is not None:
        # better way of accessing the database name?
        return str(db).split("/")[-1]

    storage_path = create_path(storage)
    if isinstance(storage_path, LocalPathClasses):
        name = storage_path.stem
    else:
        name = storage_path._url.netloc
    name = name.lower()

    return name



import lnschema_bionty as lb
import pytest

import lamindb as ln

lb.settings.auto_save_parents = False


adata = ln.dev.datasets.anndata_with_obs()
# add another column
adata.obs["cell_type_from_expert"] = adata.obs["cell_type"]
adata.obs.loc["obs0", "cell_type_from_expert"] = "B cell"


def test_labels_add():
    label = ln.ULabel(name="Experiment 1")
    file = ln.File(adata, description="test")
    file.save()
    experiment = ln.Feature(name="experiment", type="category")
    with pytest.raises(ValueError) as error:
        file.labels.add("experiment_1", experiment)
    assert (
        error.exconly()
        == "ValueError: Please pass a record (a `Registry` object), not a string, e.g.,"
        " via: label = ln.ULabel(name='experiment_1')"
    )
    with pytest.raises(ln.dev.exceptions.ValidationError) as error:
        file.labels.add(label, experiment)
    assert "not validated. If it looks correct: record.save()" in error.exconly()
    label.save()
    with pytest.raises(TypeError) as error:
        file.labels.add(label, "experiment 1")
    with pytest.raises(ln.dev.exceptions.ValidationError) as error:
        file.labels.add(label, feature=experiment)
    assert (
        error.exconly()
        == "lamindb.dev.exceptions.ValidationError: Feature not validated. If it looks"
        " correct: ln.Feature(name='experiment', type='category',"
        " registries='core.ULabel').save()"
    )
    experiment.save()
    file.labels.add(label, feature=experiment)
    # check that the feature was updated with registries = "core.ULabel"
    feature = ln.Feature.filter(name="experiment").one()
    assert feature.registries == "core.ULabel"
    with pytest.raises(TypeError):
        experiments = file.labels.get("experiment")
    # check that the label is there, it's exactly one label with name "Experiment 1"
    experiments = file.labels.get(experiment)
    assert experiments.get().name == "Experiment 1"

    # try adding the same label again, nothing should happen
    file.labels.add(label, feature=experiment)
    # check that the label is there, it's exactly one label with name "Experiment 1"
    experiments = file.labels.get(experiment)
    assert experiments.get().name == "Experiment 1"

    feature_set_n1 = ln.FeatureSet.filter(features__name="experiment").one()

    # running from_values to load validated label records under the hood
    experiment = ln.Feature(
        name="experiment_with_reg", type="category", registries=[ln.ULabel]
    )
    experiment.save()
    ln.ULabel(name="Experiment 2").save()
    file.labels.add("Experiment 2", experiment)
    experiments = file.labels.get(experiment)
    assert experiments.get().name == "Experiment 2"

    # now, try adding a new label for a new feature, extending the feature set
    project = ln.ULabel(name="project 1")
    project.save()
    ln.Feature(name="project", type="category").save()
    features = ln.Feature.lookup()
    file.labels.add(project, feature=features.project)
    # check that the label is there, it's exactly one label with name "Experiment 1"
    projects = file.labels.get(features.project)
    assert projects.get().name == "project 1"

    # here, we test that feature_set_n1 was removed because it was no longer
    # linked to any file
    feature_set_n2 = ln.FeatureSet.filter(features__name="experiment").one()
    assert feature_set_n1.id != feature_set_n2.id
    assert file.feature_sets.get() == feature_set_n2

    file.delete(storage=True)
    ln.Feature.filter().all().delete()
    ln.ULabel.filter().all().delete()
    ln.FeatureSet.filter().all().delete()


def test_add_labels_using_anndata():
    species = lb.Species.from_bionty(name="mouse")
    cell_types = [lb.CellType(name=name) for name in adata.obs["cell_type"].unique()]
    ln.save(cell_types)
    cell_types_from_expert = [
        lb.CellType(name=name) for name in adata.obs["cell_type_from_expert"].unique()
    ]
    ln.save(cell_types_from_expert)
    actual_tissues = [lb.Tissue(name=name) for name in adata.obs["tissue"].unique()]
    organoid = ln.ULabel(name="organoid")
    tissues = actual_tissues + [organoid]
    ln.save(tissues)

    lb.settings.species = "human"
    lb.settings.auto_save_parents = False

    # clean up DB state
    species_feature = ln.Feature.filter(name="species").one_or_none()
    if species_feature is not None:
        species_feature.delete()
    file = ln.File.filter(description="Mini adata").one_or_none()
    if file is not None:
        file.delete(storage=True)
    ln.FeatureSet.filter().all().delete()

    # try to construct without registering metadata features
    file = ln.File.from_anndata(
        adata, description="Mini adata", field=lb.Gene.ensembl_gene_id
    )
    assert "obs" not in file._feature_sets
    # add feature set without saving file
    feature_name_feature = ln.Feature(
        name="feature name", type="category", registries="core.ULabel"
    )
    feature_name_feature.save()
    feature_set = ln.FeatureSet(features=[feature_name_feature])
    with pytest.raises(ValueError) as error:
        file.features.add_feature_set(feature_set, slot="random")
    assert (
        error.exconly()
        == "ValueError: Please save the file or dataset before adding a feature set!"
    )

    # now register features we want to validate
    # (we are not interested in cell_type_id, here)
    ln.save(
        ln.Feature.from_df(
            adata.obs[["cell_type", "tissue", "cell_type_from_expert", "disease"]]
        )
    )
    file = ln.File.from_anndata(
        adata, description="Mini adata", field=lb.Gene.ensembl_gene_id
    )
    ln.Feature(name="species", type="category", registries="bionty.Species").save()
    features = ln.Feature.lookup()
    with pytest.raises(ValueError) as error:
        file.labels.add(species, feature=features.species)
    assert (
        error.exconly()
        == "ValueError: Please save the file/dataset before adding a label!"
    )
    file.save()

    # check the basic construction of the feature set based on obs
    feature_set_obs = file.feature_sets.filter(
        registry="core.Feature", filefeatureset__slot="obs"
    ).one()
    assert feature_set_obs.n == 4
    assert "species" not in feature_set_obs.features.list("name")

    # now, we add species and run checks
    with pytest.raises(ln.dev.exceptions.ValidationError):
        file.labels.add(species, feature=features.species)
    species.save()
    file.labels.add(species, feature=features.species)
    feature = ln.Feature.filter(name="species").one()
    assert feature.type == "category"
    assert feature.registries == "bionty.Species"
    feature_set_obs = file.feature_sets.filter(
        registry="core.Feature", filefeatureset__slot="obs"
    ).one()
    assert feature_set_obs.n == 4
    feature_set_ext = file.feature_sets.filter(
        registry="core.Feature", filefeatureset__slot="external"
    ).one()
    assert feature_set_ext.n == 1
    assert "species" in feature_set_ext.features.list("name")

    # now we add cell types & tissues and run checks
    ln.Feature(name="cell_type", type="category").save()
    ln.Feature(name="cell_type_from_expert", type="category").save()
    ln.Feature(name="tissue", type="category").save()
    file.labels.add(cell_types, feature=features.cell_type)
    file.labels.add(cell_types_from_expert, feature=features.cell_type_from_expert)
    file.labels.add(tissues, feature=features.tissue)
    feature = ln.Feature.filter(name="cell_type").one()
    assert feature.registries == "bionty.CellType"
    feature = ln.Feature.filter(name="cell_type_from_expert").one()
    assert feature.registries == "bionty.CellType"
    feature = ln.Feature.filter(name="tissue").one()
    assert feature.registries == "bionty.Tissue|core.ULabel"
    diseases = [ln.ULabel(name=name) for name in adata.obs["disease"].unique()]
    ln.save(diseases)
    file.labels.add(diseases, feature=features.disease)
    df = file.features["obs"].df()
    assert set(df["name"]) == {
        "cell_type",
        "disease",
        "tissue",
        "cell_type_from_expert",
    }
    assert set(df["type"]) == {
        "category",
    }
    assert set(df["registries"]) == {
        "bionty.CellType",
        "core.ULabel",
        "bionty.Tissue|core.ULabel",
    }

    # now, let's add another feature to ext
    experiment_1 = ln.ULabel(name="experiment_1")
    experiment_1.save()
    ln.Feature(name="experiment", type="category").save()
    features = ln.Feature.lookup()
    file.labels.add(experiment_1, feature=features.experiment)
    df = file.features["external"].df()
    assert set(df["name"]) == {
        "species",
        "experiment",
    }
    assert set(df["type"]) == {
        "category",
    }

    assert set(file.labels.get(features.experiment).list("name")) == {"experiment_1"}
    assert set(file.labels.get(features.disease).list("name")) == {
        "chronic kidney disease",
        "Alzheimer disease",
        "liver lymphoma",
        "cardiac ventricle disorder",
    }
    assert set(file.labels.get(features.species).list("name")) == {"mouse"}
    assert set(file.labels.get(features.tissue)["bionty.Tissue"].list("name")) == {
        "liver",
        "heart",
        "kidney",
        "brain",
    }
    assert set(file.labels.get(features.tissue)["core.ULabel"].list("name")) == {
        "organoid",
    }
    # currently, we can't stratify the two cases below
    assert set(file.labels.get(features.cell_type).list("name")) == {
        "T cell",
        "my new cell type",
        "hepatocyte",
        "hematopoietic stem cell",
        "B cell",
    }
    assert set(file.labels.get(features.cell_type, flat_names=True)) == {
        "T cell",
        "my new cell type",
        "hepatocyte",
        "hematopoietic stem cell",
        "B cell",
    }
    assert set(file.labels.get(features.cell_type_from_expert).list("name")) == {
        "T cell",
        "my new cell type",
        "hepatocyte",
        "hematopoietic stem cell",
        "B cell",
    }

    assert set(df["registries"]) == {"bionty.Species", "core.ULabel"}
    assert experiment_1 in file.ulabels.all()

    # call describe
    file.describe()

    # clean up
    ln.Feature.filter(name="species").one().delete()
    ln.File.filter(description="Mini adata").one().delete(storage=True)
    ln.FeatureSet.filter().all().delete()
    feature_name_feature.delete()
    lb.CellType.filter().all().delete()
    lb.Tissue.filter().all().delete()
    lb.Disease.filter().all().delete()
    ln.ULabel.filter().all().delete()


def test_labels_get():
    ln.dev.datasets.file_mini_csv()
    file = ln.File("mini.csv", description="test")
    # feature doesn't exist
    with pytest.raises(TypeError):
        file.labels.get("x")
    # no linked labels
    feature_name_feature = ln.Feature(name="feature name", type="category")
    feature_name_feature.save()
    feature_set = ln.FeatureSet(features=[feature_name_feature])
    feature_set.save()
    file.save()
    assert str(file.features) == "no linked features"
    file.features.add_feature_set(feature_set, slot="random")
    assert file.feature_sets["random"] == feature_set
    file.delete(storage=True)
    feature_set.delete()
    feature_name_feature.delete()



# this tries to load a persistent instance created in a public s3 bucket
# with s3:GetObject and s3:ListBucket policies enabled for all
# the bucket is s3://lamin-site-assets

from uuid import UUID

import lamindb_setup as ln_setup
from lamindb_setup.dev._hub_client import connect_hub_with_auth
from lamindb_setup.dev._hub_crud import select_instance_by_owner_name


def test_load_persistent_instance():
    assert ln_setup.dev.upath.AWS_CREDENTIALS_PRESENT is None
    ln_setup.login("static-testuser1@lamin.ai", password="static-testuser1-password")
    ln_setup.load("static-testuser1/static-testinstance1")
    hub = connect_hub_with_auth()
    instance = select_instance_by_owner_name(
        owner=ln_setup.settings.instance.owner,
        name=ln_setup.settings.instance.name,
        client=hub,
    )
    hub.auth.sign_out()
    assert ln_setup.settings.instance.id == UUID(instance["id"])
    assert not ln_setup.dev.upath.AWS_CREDENTIALS_PRESENT
    ln_setup.close()



"""Developer API.

Paths and file system:

.. autosummary::
   :toctree:

   upath

Settings:

.. autosummary::
   :toctree:

   UserSettings
   InstanceSettings
   StorageSettings

"""
from . import upath, django
from ._deprecated import deprecated
from ._docs import doc_args
from ._settings_instance import InstanceSettings
from ._settings_storage import StorageSettings
from ._settings_user import UserSettings
from .django import MISSING_MIGRATIONS_WARNING



import os

import lamindb_setup as ln_setup


def test_user_exists():
    # cannot test this on staging because of email rate limit right now
    if os.getenv("LAMIN_ENV") == "prod":
        # test user exists
        assert ln_setup.signup("testuser1@lamin.ai") == "user-exists"



from ._settings_instance import InstanceSettings


def write_bionty_sources(isettings: InstanceSettings) -> None:
    """Write bionty sources to BiontySource table."""
    if "bionty" not in isettings.schema:
        return None
    import shutil
    from bionty.dev._handle_sources import (
        CURRENT_SOURCES,
        LAMINDB_SOURCES,
        LOCAL_SOURCES,
        parse_sources_yaml,
    )

    shutil.copy(CURRENT_SOURCES, LAMINDB_SOURCES)

    import bionty as bt
    from lnschema_bionty.models import BiontySource

    all_sources = parse_sources_yaml(LOCAL_SOURCES)
    all_sources_dict = all_sources.to_dict(orient="records")

    currently_used = (
        bt.display_currently_used_sources()
        .reset_index()
        .set_index(["entity", "species"])
    )

    all_records = []
    for kwargs in all_sources_dict:
        act = currently_used.loc[(kwargs["entity"], kwargs["species"])].to_dict()
        if (act["source"] == kwargs["source"]) and (
            act["version"] == kwargs["version"]
        ):
            kwargs["currently_used"] = True

        record = BiontySource(**kwargs)
        all_records.append(record)

    BiontySource.objects.bulk_create(all_records, ignore_conflicts=True)


def load_bionty_sources(isettings: InstanceSettings):
    """Write currently_used bionty sources to LAMINDB_VERSIONS_PATH in bionty."""
    if "bionty" not in isettings.schema:
        return None
    from bionty.dev._handle_sources import (
        LAMINDB_SOURCES,
        parse_currently_used_sources,
    )
    from bionty.dev._io import write_yaml
    from lnschema_bionty.models import BiontySource

    active_records = BiontySource.objects.filter(currently_used=True).all().values()
    write_yaml(parse_currently_used_sources(active_records), LAMINDB_SOURCES)


def delete_bionty_sources_yaml():
    """Delete LAMINDB_SOURCES_PATH in bionty."""
    try:
        from bionty.dev._handle_sources import LAMINDB_SOURCES

        LAMINDB_SOURCES.unlink(missing_ok=True)
    except ModuleNotFoundError:
        pass



from typing import Iterable, List, Optional, Type, Union

import numpy as np
import pandas as pd
from lamin_utils import logger
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Feature, FeatureSet, Modality, Registry, ids
from lnschema_core.types import FieldAttr, ListLike

from lamindb._utils import attach_func_to_class_method
from lamindb.dev.hashing import hash_set

from . import _TESTING
from ._feature import convert_numpy_dtype_to_lamin_feature_type
from ._registry import init_self_from_db
from ._save import bulk_create
from .dev._priors import NUMBER_TYPE


def get_related_name(features_type: Registry):
    candidates = [
        field.related_name
        for field in FeatureSet._meta.related_objects
        if field.related_model == features_type
    ]
    if not candidates:
        raise ValueError(
            f"Can't create feature sets from {features_type.__name__} because it's not"
            " related to it!\nYou need to create a link model between FeatureSet and"
            " your Registry in your custom schema.\nTo do so, add a"
            " line:\nfeature_sets = models.ManyToMany(FeatureSet,"
            " related_name='mythings')\n"
        )
    return candidates[0]


def validate_features(features: List[Registry]) -> Registry:
    """Validate and return feature type."""
    try:
        if len(features) == 0:
            raise ValueError("Provide list of features with at least one element")
    except TypeError:
        raise ValueError("Please pass a ListLike of features, not a single feature")
    if not hasattr(features, "__getitem__"):
        raise TypeError("features has to be list-like")
    if not isinstance(features[0], Registry):
        raise TypeError(
            "features has to store feature records! use .from_values() otherwise"
        )
    feature_types = set([feature.__class__ for feature in features])
    if len(feature_types) > 1:
        raise TypeError("feature_set can only contain a single type")
    for feature in features:
        if feature._state.adding:
            raise ValueError("Can only construct feature sets from validated features")
    return next(iter(feature_types))  # return value in set of cardinality 1


def __init__(self, *args, **kwargs):
    if len(args) == len(self._meta.concrete_fields):
        super(FeatureSet, self).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: features")
    features: Iterable[Registry] = kwargs.pop("features") if len(args) == 0 else args[0]
    type: Optional[Union[type, str]] = kwargs.pop("type") if "type" in kwargs else None
    modality: Optional[Modality] = (
        kwargs.pop("modality") if "modality" in kwargs else None
    )
    name: Optional[str] = kwargs.pop("name") if "name" in kwargs else None
    if len(kwargs) > 0:
        raise ValueError(
            "Only features, type, modality, name are valid keyword arguments"
        )
    # now code
    features_registry = validate_features(features)
    if type is None:
        type = None if features_registry == Feature else NUMBER_TYPE
    n_features = len(features)
    features_hash = hash_set({feature.id for feature in features})
    feature_set = FeatureSet.filter(hash=features_hash).one_or_none()
    if feature_set is not None:
        logger.success(f"loaded: {feature_set}")
        init_self_from_db(self, feature_set)
        return None
    else:
        hash = features_hash
    self._features = (get_related_name(features_registry), features)
    if modality is not None:
        if not isinstance(modality, Modality):
            raise TypeError("modality needs to be of type Modality")
        if modality._state.adding:
            raise ValueError(
                "unvalidated modality, save to registry if you're sure it is correct:"
                f" {modality}"
            )
    super(FeatureSet, self).__init__(
        id=ids.base62_20(),
        name=name,
        type=get_type_str(type),
        n=n_features,
        modality=modality,
        registry=features_registry.__get_name_with_schema__(),
        hash=hash,
    )


@doc_args(FeatureSet.save.__doc__)
def save(self, *args, **kwargs) -> None:
    """{}"""
    super(FeatureSet, self).save(*args, **kwargs)
    if hasattr(self, "_features"):
        related_name, records = self._features
        # if values are stored in their dedicated table, we can bulk_create
        if related_name != "features":
            bulk_create(records)
        else:
            # otherwise, we currently need to save one by one
            for record in records:
                record.save()
        getattr(self, related_name).set(records)


def get_type_str(type: Optional[Union[Type, str]]) -> Optional[str]:
    if type is not None:
        type_str = type.__name__ if not isinstance(type, str) else type
    else:
        type_str = None
    if type == "int" or type == "float":
        type_str = NUMBER_TYPE
    return type_str


@classmethod  # type:ignore
@doc_args(FeatureSet.from_values.__doc__)
def from_values(
    cls,
    values: ListLike,
    field: FieldAttr = Feature.name,
    type: Optional[Union[Type, str]] = None,
    name: Optional[str] = None,
    modality: Optional[Modality] = None,
    **kwargs,
) -> Optional["FeatureSet"]:
    """{}"""
    if not isinstance(field, FieldAttr):
        raise TypeError(
            "Argument `field` must be a Registry field, e.g., `Feature.name`"
        )
    if len(values) == 0:
        raise ValueError("Provide a list of at least one value")
    registry = field.field.model
    if registry != Feature and type is None:
        type = NUMBER_TYPE
        logger.debug(f"setting feature set to '{NUMBER_TYPE}'")
    validated = registry.validate(values, field=field)
    if validated.sum() == 0:
        logger.warning("no validated features, skip creating feature set")
        return None
    validated_values = np.array(values)[validated]
    validated_features = registry.from_values(validated_values, field=field, **kwargs)
    feature_set = FeatureSet(
        features=validated_features,
        name=name,
        modality=modality,
        type=get_type_str(type),
    )
    return feature_set


@classmethod  # type:ignore
@doc_args(FeatureSet.from_df.__doc__)
def from_df(
    cls,
    df: "pd.DataFrame",
    field: FieldAttr = Feature.name,
    name: Optional[str] = None,
    modality: Optional[Modality] = None,
) -> Optional["FeatureSet"]:
    """{}"""
    registry = field.field.model
    validated = registry.validate(df.columns, field=field)
    if validated.sum() == 0:
        logger.warning("no validated features, skip creating feature set")
        return None
    if registry == Feature:
        validated_features = Feature.from_df(df.loc[:, validated])
        feature_set = FeatureSet(
            validated_features, name=name, type=None, modality=modality
        )
    else:
        dtypes = [col.dtype for (_, col) in df.loc[:, validated].items()]
        if len(set(dtypes)) != 1:
            raise ValueError(f"data types are heterogeneous: {set(dtypes)}")
        type = convert_numpy_dtype_to_lamin_feature_type(dtypes[0])
        validated_features = registry.from_values(df.columns[validated], field=field)
        feature_set = FeatureSet(
            features=validated_features,
            name=name,
            modality=modality,
            type=get_type_str(type),
        )
    return feature_set


METHOD_NAMES = [
    "__init__",
    "from_values",
    "from_df",
    "save",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(FeatureSet, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, FeatureSet, globals())



import lamindb as ln
from lamindb.dev._run_context import get_transform_kwargs_from_nbproject


def test_track_with_multi_parents():
    parent1 = ln.Transform(name="parent 1")
    parent1.save()
    parent2 = ln.Transform(name="parent 2")
    parent2.save()
    child = ln.Transform(name="Child")
    child.save()
    child.parents.set([parent1, parent2])
    ln.track(child, reference="my address", reference_type="url")
    # unset to remove side effects
    ln.dev.run_context.run = None
    ln.dev.run_context.transform = None


def test_track_with_reference():
    transform = ln.Transform(name="test")
    ln.track(transform, reference="my address", reference_type="url")
    assert ln.dev.run_context.run.reference == "my address"
    assert ln.dev.run_context.run.reference_type == "url"
    # unset to remove side effects
    ln.dev.run_context.run = None
    ln.dev.run_context.transform = None


def test_track_notebook_colab():
    notebook_path = "/fileId=1KskciVXleoTeS_OGoJasXZJreDU9La_l"
    # pretend we're running from ipython
    ln.dev._run_context.is_run_from_ipython = True
    ln.dev.run_context._track(notebook_path=notebook_path)
    ln.dev._run_context.is_run_from_ipython = False


def test_track_from_nbproject():
    title = "nbproject title"
    transform, id, version, name, old_version_of = get_transform_kwargs_from_nbproject(
        nbproject_id="NJvdsWWbJlZS", nbproject_version="0", nbproject_title=title
    )
    assert transform is None
    assert id == "NJvdsWWbJlZSz8"
    assert version == "0"
    assert name == title
    assert old_version_of is None
    ln.Transform(id=id, version=version, name=name).save()
    transform, id, version, name, old_version_of = get_transform_kwargs_from_nbproject(
        nbproject_id="NJvdsWWbJlZS", nbproject_version="0", nbproject_title=title
    )
    assert transform is not None
    transform, id, version, name, old_version_of = get_transform_kwargs_from_nbproject(
        nbproject_id="NJvdsWWbJlZS", nbproject_version="1", nbproject_title=title
    )
    assert transform is None
    assert id.startswith("NJvdsWWbJlZS")
    assert version == "1"
    assert name == title
    assert old_version_of is not None



# see the overlap with connector.py in laminhub-rest
import os
from typing import Optional
from lamin_utils import logger
from supabase.lib.client_options import ClientOptions
from urllib.request import urlretrieve
from supabase import create_client, Client
from pydantic import BaseSettings
from postgrest import APIError as PostgrestAPIError
from gotrue.errors import AuthUnknownError


class Connector(BaseSettings):
    url: str
    key: str


def load_fallback_connector() -> Connector:
    url = "https://lamin-site-assets.s3.amazonaws.com/connector.env"
    connector_file, _ = urlretrieve(url)
    connector = Connector(_env_file=connector_file)
    return connector


PROD_URL = "https://hub.lamin.ai"
PROD_KEY = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6ImxhZXNhdW1tZHlkbGxwcGdmY2h1Iiwicm9sZSI6ImFub24iLCJpYXQiOjE2NTY4NDA1NTEsImV4cCI6MTk3MjQxNjU1MX0.WUeCRiun0ExUxKIv5-CtjF6878H8u26t0JmCWx3_2-c"  # noqa


class Environment:
    def __init__(self, fallback: bool = False):
        lamin_env = os.getenv("LAMIN_ENV")
        if lamin_env is None:
            lamin_env = "prod"
        # set public key
        if lamin_env == "prod":
            if not fallback:
                url = PROD_URL
                key = PROD_KEY
            else:
                connector = load_fallback_connector()
                url = connector.url
                key = connector.key
        elif lamin_env == "staging":
            url = "https://amvrvdwndlqdzgedrqdv.supabase.co"
            key = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6ImFtdnJ2ZHduZGxxZHpnZWRycWR2Iiwicm9sZSI6ImFub24iLCJpYXQiOjE2NzcxNTcxMzMsImV4cCI6MTk5MjczMzEzM30.Gelt3dQEi8tT4j-JA36RbaZuUvxRnczvRr3iyRtzjY0"  # noqa
        else:
            url = os.environ["SUPABASE_API_URL"]
            key = os.environ["SUPABASE_ANON_KEY"]
        self.lamin_env: str = lamin_env
        self.supabase_api_url: str = url
        self.supabase_anon_key: str = key


# runs ~0.5s
def connect_hub(
    fallback_env: bool = False, client_options: ClientOptions = ClientOptions()
) -> Client:
    env = Environment(fallback=fallback_env)
    return create_client(env.supabase_api_url, env.supabase_anon_key, client_options)


def connect_hub_with_auth(
    fallback_env: bool = False, renew_token: bool = False
) -> Client:
    from lamindb_setup import settings

    hub = connect_hub(fallback_env=fallback_env)
    if renew_token:
        settings.user.access_token = get_access_token(
            settings.user.email, settings.user.password
        )
    hub.postgrest.auth(settings.user.access_token)
    return hub


# runs ~0.5s
def get_access_token(email: Optional[str] = None, password: Optional[str] = None):
    hub = connect_hub()
    try:
        auth_response = hub.auth.sign_in_with_password(
            {
                "email": email,
                "password": password,
            }
        )
        return auth_response.session.access_token
    except Exception as e:
        logger.error(
            f"Could not authenticate with email {email} and password {password}"
        )
        raise e
    finally:
        hub.auth.sign_out()


def call_with_fallback_auth(
    callable,
    **kwargs,
):
    for renew_token, fallback_env in [(False, False), (True, False), (False, True)]:
        try:
            client = connect_hub_with_auth(
                renew_token=renew_token, fallback_env=fallback_env
            )
            result = callable(**kwargs, client=client)
            break
        except (PostgrestAPIError, AuthUnknownError) as e:
            if fallback_env:
                raise e
        finally:
            client.auth.sign_out()
    return result


def call_with_fallback(
    callable,
    **kwargs,
):
    for fallback_env in [False, True]:
        try:
            client = connect_hub(fallback_env=fallback_env)
            result = callable(**kwargs, client=client)
            break
        except AuthUnknownError as e:
            if fallback_env:
                raise e
        finally:
            # in case there was sign in
            client.auth.sign_out()
    return result



import importlib
import inspect
from typing import List, Optional

from IPython.display import display
from lamin_utils import colors, logger
from lamindb_setup import settings
from lamindb_setup._init_instance import get_schema_module_name
from lnschema_core import Registry


def view(
    n: int = 7, schema: Optional[str] = None, registries: Optional[List[str]] = None
):
    """View data.

    Args:
        n: Display the last `n` rows of a table.
        schema: Schema module to view. Default's to
            `None` and displays all schema modules.
        registries: List of Registry names. Defaults to
            `None` and lists all ORMs.

    Examples:
        >>> ln.view()
    """
    if schema is not None:
        schema_names = [schema]
    else:
        schema_names = ["core"] + list(settings.instance.schema)

    for schema_name in schema_names:
        schema_module = importlib.import_module(get_schema_module_name(schema_name))

        all_registries = {
            orm
            for orm in schema_module.__dict__.values()
            if inspect.isclass(orm)
            and issubclass(orm, Registry)
            and orm.__name__ != "Registry"
        }
        if registries is not None:
            filtered_registries = {
                orm for orm in all_registries if orm.__name__ in registries
            }
        else:
            filtered_registries = all_registries
        if len(schema_names) > 1:
            section = f"* module: {colors.green(colors.bold(schema_name))} *"
            section_no_color = f"* module: {schema_name} *"
            logger.print("*" * len(section_no_color))
            logger.print(section)
            logger.print("*" * len(section_no_color))
        for orm in sorted(filtered_registries, key=lambda x: x.__name__):
            if hasattr(orm, "updated_at"):
                df = orm.filter().order_by("-updated_at")[:n].df()
            else:
                # need to adjust in the future
                df = orm.filter().df().iloc[-n:]
            if df.shape[0] > 0:
                logger.print(colors.blue(colors.bold(orm.__name__)))
                display(df)



from typing import Dict, Iterable, Optional, Tuple, Union

import anndata as ad
import pandas as pd
from lamin_utils import logger
from lamindb_setup.dev._docs import doc_args
from lnschema_core import Modality
from lnschema_core.models import Dataset, Feature, FeatureSet
from lnschema_core.types import AnnDataLike, DataLike, FieldAttr

from lamindb._utils import attach_func_to_class_method
from lamindb.dev._data import _track_run_input
from lamindb.dev.storage._backed_access import AnnDataAccessor, BackedAccessor
from lamindb.dev.versioning import get_ids_from_old_version, init_id

from . import _TESTING, File, Run
from ._file import parse_feature_sets_from_anndata
from ._registry import init_self_from_db
from .dev._data import (
    add_transform_to_kwargs,
    get_run,
    save_feature_set_links,
    save_transform_run_feature_sets,
)
from .dev.hashing import hash_set


def __init__(
    dataset: Dataset,
    *args,
    **kwargs,
):
    if len(args) == len(dataset._meta.concrete_fields):
        super(Dataset, dataset).__init__(*args, **kwargs)
        return None
    # now we proceed with the user-facing constructor
    if len(args) > 1:
        raise ValueError("Only one non-keyword arg allowed: data")
    data: Union[pd.DataFrame, ad.AnnData, File, Iterable[File]] = (
        kwargs.pop("data") if len(args) == 0 else args[0]
    )
    name: Optional[str] = kwargs.pop("name") if "name" in kwargs else None
    description: Optional[str] = (
        kwargs.pop("description") if "description" in kwargs else None
    )
    reference: Optional[str] = (
        kwargs.pop("reference") if "reference" in kwargs else None
    )
    reference_type: Optional[str] = (
        kwargs.pop("reference_type") if "reference_type" in kwargs else None
    )
    run: Optional[Run] = kwargs.pop("run") if "run" in kwargs else None
    is_new_version_of: Optional[Dataset] = (
        kwargs.pop("is_new_version_of") if "is_new_version_of" in kwargs else None
    )
    initial_version_id: Optional[str] = (
        kwargs.pop("initial_version_id") if "initial_version_id" in kwargs else None
    )
    version: Optional[str] = kwargs.pop("version") if "version" in kwargs else None
    feature_sets: Dict[str, FeatureSet] = (
        kwargs.pop("feature_sets") if "feature_sets" in kwargs else {}
    )
    if not len(kwargs) == 0:
        raise ValueError(
            f"Only data, name, run, description, reference, reference_type can be passed, you passed: {kwargs}"  # noqa
        )

    if is_new_version_of is None:
        provisional_id = init_id(version=version, n_full_id=20)
    else:
        if not isinstance(is_new_version_of, Dataset):
            raise TypeError("is_new_version_of has to be of type ln.Dataset")
        provisional_id, initial_version_id, version = get_ids_from_old_version(
            is_new_version_of, version, n_full_id=20
        )
        if name is None:
            name = is_new_version_of.name
    if version is not None:
        if initial_version_id is None:
            logger.info(
                "initializing versioning for this dataset! create future versions of it"
                " using ln.Dataset(..., is_new_version_of=old_dataset)"
            )

    run = get_run(run)
    # there are exactly two ways of creating a Dataset object right now
    # using exactly one file or using more than one file
    # init file
    if isinstance(data, (pd.DataFrame, ad.AnnData, File)):
        files = None
        if isinstance(data, File):
            file = data
            if file._state.adding:
                raise ValueError("Save file before creating dataset!")
            if not feature_sets:
                feature_sets = file.features._feature_set_by_slot
            else:
                if len(file.features._feature_set_by_slot) > 0:
                    logger.info("overwriting feature sets linked to file")
        else:
            log_hint = True if feature_sets is None else False
            file_is_new_version_of = (
                is_new_version_of.file if is_new_version_of is not None else None
            )
            file = File(
                data,
                run=run,
                description="tmp",
                log_hint=log_hint,
                version=version,
                is_new_version_of=file_is_new_version_of,
            )
        hash = file.hash  # type: ignore
        provisional_id = file.id  # type: ignore
        file.description = f"See dataset {provisional_id}"  # type: ignore
        file._feature_sets = feature_sets
    # init files
    else:
        file = None
        if hasattr(data, "__getitem__"):
            assert isinstance(data[0], File)  # type: ignore
            files = data
            hash, feature_sets = from_files(files)  # type: ignore
        else:
            raise ValueError("Only DataFrame, AnnData and iterable of File is allowed")
    existing_dataset = Dataset.filter(hash=hash).one_or_none()
    if existing_dataset is not None:
        logger.warning(f"returning existing dataset with same hash: {existing_dataset}")
        init_self_from_db(dataset, existing_dataset)
        for slot, feature_set in dataset.features._feature_set_by_slot.items():
            if slot in feature_sets:
                if not feature_sets[slot] == feature_set:
                    dataset.feature_sets.remove(feature_set)
                    logger.warning(f"removing feature set: {feature_set}")
    else:
        kwargs = {}
        add_transform_to_kwargs(kwargs, run)
        super(Dataset, dataset).__init__(
            id=provisional_id,
            name=name,
            description=description,
            reference=reference,
            reference_type=reference_type,
            file=file,
            hash=hash,
            run=run,
            version=version,
            initial_version_id=initial_version_id,
            **kwargs,
        )
    dataset._files = files
    dataset._feature_sets = feature_sets
    # register provenance
    if file is not None:
        _track_run_input(file, run=run)
    elif files is not None:
        for file in files:
            _track_run_input(file, run=run)
    # there is not other possibility


@classmethod  # type: ignore
@doc_args(Dataset.from_df.__doc__)
def from_df(
    cls,
    df: "pd.DataFrame",
    field: FieldAttr = Feature.name,
    name: Optional[str] = None,
    description: Optional[str] = None,
    run: Optional[Run] = None,
    modality: Optional[Modality] = None,
    reference: Optional[str] = None,
    reference_type: Optional[str] = None,
) -> "Dataset":
    """{}"""
    feature_set = FeatureSet.from_df(df, field=field, modality=modality)
    if feature_set is not None:
        feature_sets = {"columns": feature_set}
    else:
        feature_sets = {}
    dataset = Dataset(
        data=df, name=name, run=run, description=description, feature_sets=feature_sets
    )
    return dataset


@classmethod  # type: ignore
@doc_args(Dataset.from_anndata.__doc__)
def from_anndata(
    cls,
    adata: "AnnDataLike",
    field: Optional[FieldAttr],
    name: Optional[str] = None,
    description: Optional[str] = None,
    run: Optional[Run] = None,
    modality: Optional[Modality] = None,
    reference: Optional[str] = None,
    reference_type: Optional[str] = None,
) -> "Dataset":
    """{}"""
    if isinstance(adata, File):
        assert not adata._state.adding
        assert adata.accessor == "AnnData"
        adata_parse = adata.path
    else:
        adata_parse = adata
    feature_sets = parse_feature_sets_from_anndata(adata_parse, field, modality)
    dataset = Dataset(
        data=adata,
        run=run,
        name=name,
        description=description,
        feature_sets=feature_sets,
    )
    return dataset


# internal function, not exposed to user
def from_files(files: Iterable[File]) -> Tuple[str, Dict[str, str]]:
    # assert all files are already saved
    saved = not any([file._state.adding for file in files])
    if not saved:
        raise ValueError("Not all files are yet saved, please save them")
    # query all feature sets of files
    file_ids = [file.id for file in files]
    # query all feature sets at the same time rather than making a single query per file
    feature_set_file_links = File.feature_sets.through.objects.filter(
        file_id__in=file_ids
    )
    feature_set_slots_ids = {}
    for link in feature_set_file_links:
        feature_set_slots_ids[link.slot] = link.feature_set_id
    # validate consistency of hashes
    # we do not allow duplicate hashes
    hashes = [file.hash for file in files]
    if len(hashes) != len(set(hashes)):
        seen = set()
        non_unique = [x for x in hashes if x in seen or seen.add(x)]  # type: ignore
        raise ValueError(
            "Please pass files with distinct hashes: these ones are non-unique"
            f" {non_unique}"
        )
    hash = hash_set(set(hashes))
    return hash, feature_set_slots_ids


# docstring handled through attach_func_to_class_method
def backed(
    self, is_run_input: Optional[bool] = None
) -> Union["AnnDataAccessor", "BackedAccessor"]:
    _track_run_input(self, is_run_input)
    if self.file is None:
        raise RuntimeError("Can only call backed() for datasets with a single file")
    return self.file.backed()


# docstring handled through attach_func_to_class_method
def load(self, is_run_input: Optional[bool] = None, **kwargs) -> DataLike:
    _track_run_input(self, is_run_input)
    if self.file is not None:
        return self.file.load()
    else:
        all_files = self.files.all()
        suffixes = [file.suffix for file in all_files]
        if len(set(suffixes)) != 1:
            raise RuntimeError(
                "Can only load datasets where all files have the same suffix"
            )
        # because we're tracking data flow on the dataset-level, here, we don't
        # want to track it on the file-level
        objects = [file.load(is_run_input=False) for file in all_files]
        file_ids = [file.id for file in all_files]
        if isinstance(objects[0], pd.DataFrame):
            return pd.concat(objects)
        elif isinstance(objects[0], ad.AnnData):
            return ad.concat(objects, label="file_id", keys=file_ids)


# docstring handled through attach_func_to_class_method
def delete(self, storage: Optional[bool] = None) -> None:
    super(Dataset, self).delete()
    if self.file is not None:
        self.file.delete(storage=storage)


# docstring handled through attach_func_to_class_method
def save(self, *args, **kwargs) -> None:
    if self.file is not None:
        self.file.save()
    # we don't need to save feature sets again
    save_transform_run_feature_sets(self)
    super(Dataset, self).save()
    if hasattr(self, "_files"):
        if self._files is not None and len(self._files) > 0:
            self.files.set(self._files)
    save_feature_set_links(self)


METHOD_NAMES = [
    "__init__",
    "from_anndata",
    "from_df",
    "backed",
    "load",
    "delete",
    "save",
]

if _TESTING:
    from inspect import signature

    SIGS = {
        name: signature(getattr(Dataset, name))
        for name in METHOD_NAMES
        if name != "__init__"
    }

for name in METHOD_NAMES:
    attach_func_to_class_method(name, Dataset, globals())



from lamin_utils import logger

from ._init_instance import reload_lamindb, reload_schema_modules
from ._silence_loggers import silence_loggers
from .dev._settings_store import current_instance_settings_file

_INSTANCE_NOT_SETUP_WARNING = """\
You haven't yet setup an instance: Please call `ln.setup.init()` or `ln.setup.load()`
"""


def check_instance_setup(from_lamindb: bool = False):
    if current_instance_settings_file().exists():
        silence_loggers()
        try:
            # attempt loading the settings file
            from .dev._settings_load import load_instance_settings

            isettings = load_instance_settings()

            from .dev.django import IS_SETUP, setup_django

            # this flag should probably be renamed to `from_user`
            # it will typically be invoked if lamindb is imported for use
            # but users might also import their schema modules first
            # and then want lamindb be to be available
            if from_lamindb:
                # this guarantees that ths is called exactly once
                # prior to django being setup!
                if not IS_SETUP:
                    setup_django(isettings)
                    reload_schema_modules(isettings)
                    # only now we can import lamindb
                    reload_lamindb(isettings)
                return True
            else:
                return IS_SETUP
        except Exception:
            # user will get more detailed traceback once they run the CLI
            raise RuntimeError(
                "Current instance cannot be reached, close it: `lamin close`\n"
                "Alternatively, init or load a connectable instance on the"
                " command line: `lamin load <instance>` or `lamin init <...>`"
            )
    else:
        if from_lamindb:
            logger.warning(_INSTANCE_NOT_SETUP_WARNING)
        return False



"""LaminDB's core registries."""
_schema_id = "yvzi"
_name = "core"
__version__ = "0.48.3"


from lamindb_setup import _check_instance_setup

_INSTANCE_SETUP = _check_instance_setup()

if _INSTANCE_SETUP:
    from . import ids, types
    from .models import ORM  # backward compat
    from .models import (  # type: ignore
        CanValidate,
        Dataset,
        Feature,
        FeatureSet,
        File,
        HasParents,
        Modality,
        Registry,
        Run,
        Storage,
        Transform,
        ULabel,
        User,
    )



import os


def test_entrypoint():
    exit_status = os.system("lamin --help")
    assert exit_status == 0


def test_migrate_create():
    exit_status = os.system("lamin migrate create")
    assert exit_status == 0


def test_migrate_squash():
    exit_status = os.system(
        "yes | lamin migrate squash --package-name lnschema_core --end-number 0019"
    )
    assert exit_status == 0



import base64
from pathlib import Path

from lamindb.dev.hashing import b16_to_b64, hash_file, to_b64_str


def test_compute_hash():
    files = [
        # file_content, hash, chunk_size, hash_type
        ("a", "DMF1ucDxtqgxw5niaXcmYQ", None, "md5"),
        ("b", "kutf_uauL-w61xx3dTFXjw", None, "md5"),
        ("abc", "kAFQmDzST7DWlj99KOF_cg", None, "md5"),
        ("a", "DMF1ucDxtqgxw5niaXcmYQ", 1, "md5"),
        ("b", "kutf_uauL-w61xx3dTFXjw", 1, "md5"),
        # the last case here triggers multi-chunk compute with sha1
        ("abc", "p0EbDbQEP1wS-Tw6TuBjKS", 1, "sha1-fl"),
    ]
    for content, hash, chunk_size, hash_type in files:
        filepath = Path("file_1.txt")
        filepath.write_text(content)
        computed_hash, computed_hash_type = hash_file(filepath, chunk_size=chunk_size)
        assert computed_hash == hash
        assert computed_hash_type == hash_type
        filepath.unlink()


def test_base64():
    mytest = "test".encode()
    b64_str = to_b64_str(mytest)
    b64_str_padded = f"{b64_str}=="
    assert base64.urlsafe_b64decode(b64_str_padded.encode()).hex() == mytest.hex()


def test_b16_to_b64():
    assert b16_to_b64("9b89c8c1acf79dba5b5341d1fff9806f") == "m4nIwaz3nbpbU0HR__mAbw"



import builtins
import hashlib
import re
from datetime import datetime, timezone
from pathlib import Path, PurePath
from typing import Dict, List, Optional, Tuple, Union

from lamin_utils import logger
from lamindb_setup import settings
from lamindb_setup.dev import InstanceSettings
from lnschema_core import Run, Transform
from lnschema_core.types import TransformType

from lamindb.dev.versioning import get_ids_from_old_version, init_id

from .hashing import to_b64_str

is_run_from_ipython = getattr(builtins, "__IPYTHON__", False)

msg_path_failed = (
    "failed to infer notebook path.\nfix: either track manually via"
    " `ln.track(ln.Transform(name='My notebook'))` or pass"
    " `notebook_path` to ln.track()"
)

msg_manual_init = (
    "\n(1) save your notebook!"
    "\n(2) attach metadata to the notebook by running the CLI:\n"
    "lamin track {notebook_path}"
    "\n(3) reload or re-open your notebook"
)


def save_or_load(msg):
    if logger._verbosity > 2:
        logger.success(msg)
    else:
        logger.important(
            msg.replace("loaded: ", "").replace("saved: ", "").replace("updated: ", "")
        )


class UpdateNbWithNonInteractiveEditorError(Exception):
    pass


class NotebookNotSavedError(Exception):
    pass


class NoTitleError(Exception):
    pass


def _seconds_modified(filepath):
    return datetime.now().timestamp() - Path(filepath).stat().st_mtime


def _write_notebook_meta(metadata):
    from nbproject import dev as nb_dev
    from nbproject._header import _env, _filepath

    nb_dev._frontend_commands._save_notebook(_env)
    nb = nb_dev.read_notebook(_filepath)
    nb.metadata["nbproject"] = metadata

    # write proper execution count
    header_re = re.compile(r"^[^#]*track\(", flags=re.MULTILINE)
    ccount = 0
    for cell in nb.cells:
        if cell["cell_type"] != "code":
            continue
        elif cell["execution_count"] is not None:
            ccount = cell["execution_count"]
        if header_re.match("".join(cell["source"])) is not None:
            cell["execution_count"] = ccount + 1
            break

    nb_dev.write_notebook(nb, _filepath)
    nb_dev._frontend_commands._reload_notebook(_env)


# also see lamindb-setup._notebook for related code
def reinitialize_notebook(
    transform: Transform, metadata: Optional[Dict] = None, ask_for_new_id: bool = True
) -> Tuple[Transform, Dict]:
    from nbproject import dev as nb_dev
    from nbproject._header import _filepath

    # init new_id, new_version
    new_id, new_version = transform.id, None

    if ask_for_new_id:
        response = input("Do you want to generate a new id? (y/n) ")
    else:
        response = "y"
    if response == "y":
        new_id = init_id(n_full_id=14)
    else:
        response = input("Do you want to set a new version (e.g. '1.1')? (y/n) ")
        if response == "y":
            new_version = input("Please type the version: ")
            new_id = get_ids_from_old_version(
                is_new_version_of=transform, version=new_version, n_full_id=14
            )

    nb = None
    if metadata is None:
        nb = nb_dev.read_notebook(_filepath)
        metadata = nb.metadata["nbproject"]

    # nbproject metadata needs the shorter nbproject id
    metadata["id"] = new_id[:12]
    metadata["version"] = new_version

    # here we check that responses to both inputs (for new id and version) were not 'n'
    if transform.id != new_id or transform.version != new_version:
        transform = Transform(
            id=new_id, version=new_version, type=TransformType.notebook
        )
    return transform, metadata


# from https://stackoverflow.com/questions/61901628
def get_notebook_name_colab() -> str:
    from socket import gethostbyname, gethostname  # type: ignore

    from requests import get  # type: ignore

    ip = gethostbyname(gethostname())  # 172.28.0.12
    try:
        name = get(f"http://{ip}:9000/api/sessions").json()[0]["name"]
    except Exception:
        logger.warning(
            "could not get notebook name from Google Colab, using: Notebook.ipynb"
        )
        name = "Notebook.ipynb"
    return name.rstrip(".ipynb")


def get_transform_kwargs_from_nbproject(
    nbproject_id: str, nbproject_version: str, nbproject_title: str
) -> Tuple[Optional[Transform], str, str, str, Optional[Transform]]:
    id_ext = to_b64_str(hashlib.md5(nbproject_version.encode()).digest())[:2]
    id = nbproject_id + id_ext
    version = nbproject_version
    transform = Transform.filter(
        id__startswith=nbproject_id, version=version
    ).one_or_none()
    name = nbproject_title
    old_version_of = None
    if transform is None:
        old_version_of = Transform.filter(id__startswith=nbproject_id).first()
    return transform, id, version, name, old_version_of


class run_context:
    """Global run context."""

    instance: Optional[InstanceSettings] = None
    """Current instance."""
    transform: Optional[Transform] = None
    """Current transform."""
    run: Optional[Run] = None
    """Current run."""

    # exposed to user as ln.track()
    @classmethod
    def _track(
        cls,
        transform: Optional[Transform] = None,
        *,
        new_run: Optional[bool] = None,
        reference: Optional[str] = None,
        reference_type: Optional[str] = None,
        notebook_path: Optional[str] = None,
        pypackage: Optional[Union[str, List[str]]] = None,
        editor: Optional[str] = None,
    ) -> None:
        """Track global `Transform` & `Run` for a notebook or pipeline.

        Creates or loads a :class:`~lamindb.Run` record and sets a global
        :class:`~lamindb.dev.run_context`.

        In a Jupyter notebook, call without any argument (metadata is parsed).
        If the notebook has no associated metadata ("is not initialized"),
        attempts to write metadata to disk. If it fails to so interactively, it
        will ask you to leverage the CLI.

        Args:
            transform: Can be of type `"pipeline"` or `"notebook"`
                (:class:`~lamindb.dev.types.TransformType`).
            new_run: If `False`, loads latest run of transform
                (default notebook), if True, creates new run (default pipeline).
            reference: Reference to pass to :class:`~lamindb.Run` record.
            reference_type: Reference type to pass to :class:`~lamindb.Run`
                record (e.g. "url").
            notebook_path: Filepath of notebook. Only needed if inference fails.
            pypackage: One or more python packages for which to parse versions.
            editor: Editor environment.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.

        Examples:

            If you're in a Jupyter notebook and installed lamindb with `pip
            install[jupyter]`, you can simply call:

            >>> ln.track()
            >>> ln.dev.run_context.transform
            >>> ln.dev.run_context.run

            If you'd like to track a pipeline, pass a
            :class:`~lamindb.Transform` object of `type` `"pipeline"`:

            >>> ln.Transform(name="Cell Ranger", version="2", type="pipeline").save()
            >>> transform = ln.Transform.filter(name="Cell Ranger", version="2").one()
            >>> ln.track(transform)
        """
        cls.instance = settings.instance
        import lamindb as ln

        if transform is None:
            is_tracked_notebook = False

            if is_run_from_ipython:
                try:
                    cls._track_notebook(
                        pypackage=pypackage,
                        notebook_path=notebook_path,
                        editor=editor,
                    )
                    is_tracked_notebook = True
                except Exception as e:
                    if isinstance(e, ImportError):
                        logger.warning(
                            "it looks like you are running ln.track() from a "
                            "notebook!\nplease install nbproject: pip install nbproject"
                        )
                    elif isinstance(e, UpdateNbWithNonInteractiveEditorError):
                        raise e
                    elif isinstance(e, (NotebookNotSavedError, NoTitleError)):
                        raise e
                    else:
                        logger.warning(f"automatic tracking of notebook failed: {e}")
                        raise e
                    is_tracked_notebook = False

            if not is_tracked_notebook:
                logger.warning(
                    "no automatic metadata detection, consider passing transform"
                )
                return None
        else:
            transform_exists = None
            if transform.id is not None:
                # transform has an id but unclear whether already saved
                transform_exists = Transform.filter(id=transform.id).first()
            if transform_exists is None:
                transform.save()
                save_or_load(f"saved: {transform}")
                transform_exists = transform
            else:
                save_or_load(f"loaded: {transform}")
            cls.transform = transform_exists

        if new_run is None:  # for notebooks, default to loading latest runs
            new_run = False if cls.transform.type == TransformType.notebook.value else True  # type: ignore  # noqa

        run = None
        if not new_run:  # try loading latest run by same user
            run = (
                ln.Run.filter(
                    transform=cls.transform, created_by_id=ln.setup.settings.user.id
                )
                .order_by("-created_at")
                .first()
            )
            if run is not None:  # loaded latest run
                run.run_at = datetime.now(timezone.utc)  # update run time
                run.reference = reference
                run.reference_type = reference_type
                run.save()
                save_or_load(f"loaded: {run}")

        if run is None:  # create new run
            run = ln.Run(
                transform=cls.transform,
                reference=reference,
                reference_type=reference_type,
            )
            run.save()
            save_or_load(f"saved: {run}")
        cls.run = run

        # at this point, we have a transform can display its parents if there are any
        parents = cls.transform.parents.all() if cls.transform is not None else []
        if len(parents) > 0:
            if len(parents) == 1:
                save_or_load(f"  parent transform: {parents[0]}")
            else:
                parents_formatted = "\n   - ".join([f"{parent}" for parent in parents])
                save_or_load(f"  parent transforms:\n   - {parents_formatted}")

        # only for newly intialized notebooks
        if hasattr(cls, "_notebook_meta"):
            _write_notebook_meta(cls._notebook_meta)  # type: ignore
            del cls._notebook_meta  # type: ignore

    @classmethod
    def _track_notebook(
        cls,
        *,
        notebook_path: Optional[str] = None,
        pypackage: Optional[Union[str, List[str]]] = None,
        editor: Optional[str] = None,
    ):
        """Infer Jupyter notebook metadata and create `Transform` record.

        Args:
            pypackage: One or more python packages to track.
            notebook_path: Filepath of notebook. Only needed if automatic
                inference fails.
            editor: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
        """
        import nbproject
        from nbproject.dev._jupyter_communicate import (
            notebook_path as get_notebook_path,
        )

        cls.instance = settings.instance

        metadata = None
        needs_init = False
        is_interactive = False
        reference = None
        version = None
        colab_id = None
        nbproject_id = None
        if notebook_path is None:
            path_env = None
            try:
                path_env = get_notebook_path(return_env=True)
            except Exception:
                raise RuntimeError(msg_path_failed)
            if path_env is None:
                raise RuntimeError(msg_path_failed)
            notebook_path, _env = path_env
        else:
            notebook_path, _env = notebook_path, editor
        if isinstance(notebook_path, (Path, PurePath)):
            notebook_path_str = notebook_path.as_posix()  # type: ignore
        else:
            notebook_path_str = str(notebook_path)
        if notebook_path_str.endswith("Untitled.ipynb"):
            raise RuntimeError("Please rename your notebook before tracking it")
        if notebook_path_str.startswith("/fileId="):
            # This is Google Colab!
            # google colab fileID looks like this
            # /fileId=1KskciVXleoTeS_OGoJasXZJreDU9La_l
            # we'll take the first 12 characters
            colab_id = notebook_path_str.replace("/fileId=", "")
            reference = f"colab_id: {colab_id}"
            filestem = get_notebook_name_colab()
            _env = "colab"
        else:
            try:
                metadata, needs_init, nb = nbproject.header(
                    pypackage=pypackage,
                    filepath=notebook_path_str,
                    env=_env if editor is None else editor,
                    metadata_only=True,
                )
                # this contains filepath if the header was run successfully
                from nbproject._header import _env, _filepath  # type: ignore
            except Exception as e:
                nbproject_failed_msg = (
                    "Auto-retrieval of notebook name & title failed.\n\nFixes: Either"
                    f" init on the CLI `lamin track {notebook_path_str}` or pass"
                    " transform manually `ln.track(ln.Transform(name='My"
                    " notebook'))`\n\nPlease consider pasting error at:"
                    f" https://github.com/laminlabs/nbproject/issues/new\n\n{e}"
                )
                raise RuntimeError(nbproject_failed_msg)
            try:
                from nbproject.dev._metadata_display import DisplayMeta
                from nbproject.dev._pypackage import infer_pypackages

                dm = DisplayMeta(metadata)
                logger.important(
                    "notebook imports:"
                    f" {' '.join(dm.pypackage(infer_pypackages(nb, pin_versions=True)))}"  # noqa
                )
            except Exception:
                logger.debug("inferring imported packages failed")
                pass

        if needs_init:
            if _env in ("lab", "notebook"):
                cls._notebook_meta = metadata  # type: ignore
            else:
                msg = msg_manual_init.format(notebook_path=notebook_path_str)
                raise UpdateNbWithNonInteractiveEditorError(msg)

        if _env in ("lab", "notebook"):
            # save the notebook in case that title was updated
            # but notebook not saved
            nbproject.dev._frontend_commands._save_notebook(_env)
            # check here if interactivity really works

        if metadata is not None:
            # this only executed if nbproject is used
            # check here if interactivity really works
            # timestamp should be recent due to save
            is_interactive = _seconds_modified(_filepath) < 1.5  # should be ~1 sec
            if not is_interactive and needs_init:
                msg = msg_manual_init.format(notebook_path=_filepath)
                raise UpdateNbWithNonInteractiveEditorError(msg)

            nbproject_id = metadata["id"]
            nbproject_version = metadata["version"]
            filestem = Path(_filepath).stem
            try:
                nbproject_title = nbproject.meta.live.title
            except IndexError:
                raise NotebookNotSavedError(
                    "The notebook is not saved, please save the notebook and"
                    " rerun `ln.track()`"
                )
            if nbproject_title is None:
                raise NoTitleError(
                    "Please add a title to your notebook in a markdown cell: # Title"
                )
        # colab parsing succesful
        if colab_id is not None:
            id = colab_id[:14]
            transform = Transform.filter(id=id).one_or_none()
            name = filestem
            short_name = None
            old_version_of = None
        # nbproject parsing successful
        elif nbproject_id is not None:
            (
                transform,
                id,
                version,
                name,
                old_version_of,
            ) = get_transform_kwargs_from_nbproject(
                nbproject_id, nbproject_version, nbproject_title
            )
            short_name = filestem
        # make a new transform record
        if transform is None:
            transform = Transform(
                id=id,
                version=version,
                name=name,
                short_name=short_name,
                reference=reference,
                is_new_version_of=old_version_of,
                type=TransformType.notebook,
            )
            transform.save()
            save_or_load(f"saved: {transform}")
        else:
            # check whether there was an update
            if transform.name != name or transform.short_name != short_name:
                response = input(
                    "Updated notebook name and/or title: Do you want to assign a"
                    " new id or version? (y/n)"
                )
                if response == "y":
                    if _env in ("lab", "notebook") and is_interactive:
                        transform, metadata = reinitialize_notebook(transform, metadata)
                        # only write metadata back to notebook if it actually changed!
                        # if filename or title changed, this does not merit a write!
                        # it's dangerous to write unnecessarily
                        cls._notebook_meta = metadata  # type: ignore
                    else:
                        msg = msg_manual_init.format(notebook_path=notebook_path)
                        raise UpdateNbWithNonInteractiveEditorError(msg)
                transform.name = name
                transform.short_name = filestem
                transform.save()
                if response == "y":
                    save_or_load(f"saved: {transform}")
                else:
                    save_or_load(f"updated: {transform}")
            else:
                save_or_load(f"loaded: {transform}")
        cls.transform = transform
